"""This module defines raster based geometry class
"""

import os
from typing import Union, Optional

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from shapely.geometry import MultiPolygon
from pyproj import CRS

from ocsmesh.geom.base import BaseGeom
from ocsmesh.raster import Raster


class SourceRaster:
    """Descriptor class used for referencing the source `Raster` object."""

    def __set__(self, obj, val: Union[Raster, str, os.PathLike]):

        if isinstance(val, (str, os.PathLike)):  # type: ignore[misc]
            val = Raster(val)

        if not isinstance(val, Raster):
            raise TypeError(
                f'Argument raster must be of type {Raster}, '
                f'not type {type(val)}.')
        obj.__dict__['source_raster'] = val

    def __get__(self, obj, objtype=None) -> Raster:
        return obj.__dict__['source_raster']


class RasterGeom(BaseGeom):
    """Raster based geometry.

    Create geometry based on a raster object. All the calculations
    are done on the raster image.

    Attributes
    ----------
    raster
        Reference to the source raster object
    crs
        CRS of the input raster

    Methods
    -------
    msh_t(**kwargs)
        Returns the `jigsawpy` vertex-edge representation of the geometry
    get_multipolygon(zmin=None, zmax=None)
        Returns `shapely` object representation of the geometry

    Notes
    -----
    This is the main object to use for handling DEM data. The objects
    of this class hold a reference to the original raster object and
    use that handle to extract data using `Raster` API.
    """

    _source_raster = SourceRaster()

    def __init__(
            self,
            raster: Union[Raster, str, os.PathLike],
            zmin: Optional[float] = None,
            zmax: Optional[float] = None,
        ) -> None:
        """Initialize a raster based geometry object.

        Parameters
        ----------
        raster : Raster or str or path-like
            Input raster for generating the geometry, i.e. the
            simulation domain
        zmin : float or None, default=None
            Minimum elevation cut-off for calculating polygon
        zmax : float or None, default=None
            Maximum elevation cut-off for calculating polygon
        """

        self._source_raster = raster
        self._zmin = zmin
        self._zmax = zmax

    def get_multipolygon(  # type: ignore[override]
            self,
            zmin: Optional[float] = None,
            zmax: Optional[float] = None
            ) -> MultiPolygon:
        """Returns the `shapely` representation of the geometry

        Calculates and returns the `MultiPolygon` representation of
        the geometry.

        Parameters
        ----------
        zmin : float or None, default=None
            Minimum elevation cut-off for calculating polygon, overrides
            value provided during object construction
        zmax : float or None, default=None
            Maximum elevation cut-off for calculating polygon, overrides
            value provided during object construction

        Returns
        -------
        MultiPolygon
            Calculated polygon from raster data based on the minimum
            and maximum elevations of interest.
        """

        zmin = self._zmin if zmin is None else zmin
        zmax = self._zmax if zmax is None else zmax

        if zmin is None and zmax is None:
            return MultiPolygon([self.raster.get_bbox()])

        return self.raster.get_multipolygon(zmin=zmin, zmax=zmax)


    @property
    def raster(self) -> Raster:
        """Read-only attribute for reference to the source raster"""
        return self._source_raster

    @property
    def crs(self) -> CRS:
        """Read-only attribute returning the CRS of the source raster"""
        return self.raster.crs

    def make_plot(
            self,
            ax: Optional[Axes] = None,
            show: bool = False
            ) -> Axes:
        """Create a plot from the boundaries of the geometry polygon

        Parameters
        ----------
        ax : Axes or None, default=None
            An existing axes to draw the geomtry polygon on
        show : bool, default=False
            Whether to call `pyplot.show()` after drawing the polygon

        Returns
        -------
        Axes
            The axes onto which the polygon boundaries are drawn
        """

        # TODO: Consider the ellipsoidal case. Refer to commit
        # dd087257c15692dd7d8c8e201d251ab5e66ff67f on main branch for
        # ellipsoidal ploting routing (removed).
        for polygon in self.multipolygon:
            plt.plot(*polygon.exterior.xy, color='k')
            for interior in polygon.interiors:
                plt.plot(*interior.xy, color='r')
        if show:
            plt.show()

        return plt.gca()



from collections import defaultdict
from itertools import permutations
from typing import Union, Dict, Sequence, Tuple
from functools import reduce
from multiprocessing import cpu_count, Pool
from copy import deepcopy

import jigsawpy
from jigsawpy import jigsaw_msh_t  # type: ignore[import]
from matplotlib.path import Path  # type: ignore[import]
import matplotlib.pyplot as plt  # type: ignore[import]
from matplotlib.tri import Triangulation  # type: ignore[import]
import numpy as np  # type: ignore[import]
from pyproj import CRS, Transformer  # type: ignore[import]
from scipy.interpolate import (  # type: ignore[import]
    RectBivariateSpline, griddata)
from scipy import sparse
from shapely.geometry import ( # type: ignore[import]
        Polygon, MultiPolygon,
        box, GeometryCollection, Point, MultiPoint,
        LineString, LinearRing)
from shapely.ops import polygonize, linemerge
import geopandas as gpd
import utm


ELEM_2D_TYPES = ['tria3', 'quad4', 'hexa8']

def must_be_euclidean_mesh(func):
    def decorator(mesh, *args, **kwargs):
        if mesh.mshID.lower() != 'euclidean-mesh':
            msg = f"Not implemented for mshID={mesh.mshID}"
            raise NotImplementedError(msg)
        return func(mesh, *args, **kwargs)
    return decorator


def mesh_to_tri(mesh):
    """
    mesh is a jigsawpy.jigsaw_msh_t() instance.
    """
    return Triangulation(
        mesh.vert2['coord'][:, 0],
        mesh.vert2['coord'][:, 1],
        mesh.tria3['index'])


def cleanup_isolates(mesh):

    # For triangle only (TODO: add support for other types)
    node_indexes = np.arange(mesh.vert2['coord'].shape[0])
    used_indexes = np.unique(mesh.tria3['index'])
    vert2_idxs = np.where(
        np.isin(node_indexes, used_indexes, assume_unique=True))[0]

    # Since tria simplex refers to node index which always starts from
    # 0 after removing isolate nodes we can use the map approach
    tria3 = mesh.tria3['index'].flatten()
    renum = {old: new for new, old in enumerate(np.unique(tria3))}
    tria3 = np.array([renum[i] for i in tria3])
    tria3 = tria3.reshape(mesh.tria3['index'].shape)

    mesh.vert2 = mesh.vert2.take(vert2_idxs, axis=0)
    if len(mesh.value) > 0:
        mesh.value = mesh.value.take(vert2_idxs, axis=0)
    mesh.tria3 = np.asarray(
        [(tuple(indices), mesh.tria3['IDtag'][i])
         for i, indices in enumerate(tria3)],
        dtype=jigsaw_msh_t.TRIA3_t)


def put_edge2(mesh):
    tri = Triangulation(
        mesh.vert2['coord'][:, 0],
        mesh.vert2['coord'][:, 1],
        mesh.tria3['index'])
    mesh.edge2 = np.array(
        [(edge, 0) for edge in tri.edges], dtype=jigsaw_msh_t.EDGE2_t)


def geom_to_multipolygon(mesh):
    vertices = mesh.vert2['coord']
    idx_ring_coll = index_ring_collection(mesh)
    polygon_collection = []
    for polygon in idx_ring_coll.values():
        exterior = vertices[polygon['exterior'][:, 0], :]
        interiors = []
        for interior in polygon['interiors']:
            interiors.append(vertices[interior[:, 0], :])
        polygon_collection.append(Polygon(exterior, interiors))
    return MultiPolygon(polygon_collection)


def get_boundary_segments(mesh):

    coords = mesh.vert2['coord']
    boundary_edges = get_boundary_edges(mesh)
    boundary_verts = np.unique(boundary_edges)
    boundary_coords = coords[boundary_verts]

    vert_map = {
            orig: new for new, orig in enumerate(boundary_verts)}
    new_boundary_edges = np.array(
        [vert_map[v] for v in boundary_edges.ravel()]).reshape(
                boundary_edges.shape)

    graph = sparse.lil_matrix(
            (len(boundary_verts), len(boundary_verts)))
    for vert1, vert2 in new_boundary_edges:
        graph[vert1, vert2] = 1

    n_components, labels = sparse.csgraph.connected_components(
            graph, directed=False)

    segments = []
    for i in range(n_components):
        conn_mask = np.any(np.isin(
                new_boundary_edges, np.nonzero(labels == i)),
                axis=1)
        conn_edges = new_boundary_edges[conn_mask]
        this_segment = linemerge(boundary_coords[conn_edges].tolist())
        if not this_segment.is_simple:
            # Pinched nodes also result in non-simple linestring,
            # but they can be handled gracefully, here we are looking
            # for other issues like folded elements
            test_polys = list(polygonize(this_segment))
            if not test_polys:
                raise ValueError(
                    "Mesh boundary crosses itself! Folded element(s)!")
        segments.append(this_segment)

    return segments


def get_mesh_polygons(mesh):


    # TODO: Copy mesh?
    target_mesh = mesh
    result_polys = []

    # 2-pass find, first find using polygons that intersect non-boundary
    # vertices, then from the rest of the mesh find polygons that
    # intersect any vertex
    for find_pass in range(2):


        coords = target_mesh.vert2['coord']

        if len(coords) == 0:
            continue

        boundary_edges = get_boundary_edges(target_mesh)

        lines = get_boundary_segments(target_mesh)

        poly_gen = polygonize(lines)
        polys = list(poly_gen)
        polys = sorted(polys, key=lambda p: p.area, reverse=True)


        bndry_verts = np.unique(boundary_edges)

        if find_pass == 0:
            non_bndry_verts = np.setdiff1d(
                    np.arange(len(coords)), bndry_verts)
            pnts = MultiPoint(coords[non_bndry_verts])
        else:
            pnts = MultiPoint(coords[bndry_verts])


        # NOTE: This logic requires polygons to be sorted by area
        pass_valid_polys = []
        while len(pnts.geoms) > 0:


            idx = np.random.randint(len(pnts.geoms))
            pnt = pnts.geoms[idx]

            polys_gdf = gpd.GeoDataFrame(
                {'geometry': polys, 'list_index': range(len(polys))})


            res_gdf = polys_gdf[polys_gdf.intersects(pnt)]
            if len(res_gdf) == 0:
                # How is this possible?!
                pnts = MultiPoint([*pnts.geoms[:idx], *pnts.geoms[idx + 1:]])
                if pnts.is_empty:
                    break

                continue

            poly = res_gdf.geometry.iloc[0]
            polys.pop(res_gdf.iloc[0].list_index)



            pass_valid_polys.append(poly)
            pnts = pnts.difference(poly)
            if pnts.is_empty:
                break
            if isinstance(pnts, Point):
                pnts = MultiPoint([pnts])


        result_polys.extend(pass_valid_polys)
        target_mesh = clip_mesh_by_shape(
            target_mesh,
            shape=MultiPolygon(pass_valid_polys),
            inverse=True, fit_inside=True)



    return MultiPolygon(result_polys)


def repartition_features(linestring, max_verts):
    features = []
    if len(linestring.coords) > max_verts:
        new_feat = []
        for segment in list(map(LineString, zip(
                linestring.coords[:-1],
                linestring.coords[1:]))):
            new_feat.append(segment)
            if len(new_feat) == max_verts - 1:
                features.append(linemerge(new_feat))
                new_feat = []
        if len(new_feat) != 0:
            features.append(linemerge(new_feat))
    else:
        features.append(linestring)
    return features


def transform_linestring(
    linestring: LineString,
    target_size: float,
):
    distances = [0.]
    while distances[-1] + target_size < linestring.length:
        distances.append(distances[-1] + target_size)
    distances.append(linestring.length)
    linestring = LineString([
        linestring.interpolate(distance)
        for distance in distances
        ])
    return linestring



def needs_sieve(mesh, area=None):
    areas = [polygon.area for polygon in geom_to_multipolygon(mesh)]
    if area is None:
        remove = np.where(areas < np.max(areas))[0].tolist()
    else:
        remove = []
        for idx, patch_area in enumerate(areas):
            if patch_area <= area:
                remove.append(idx)
    if len(remove) > 0:
        return True

    return False


def put_id_tags(mesh):
    # start enumerating on 1 to avoid issues with indexing on fortran models
    mesh.vert2 = np.array(
        [(coord, id+1) for id, coord in enumerate(mesh.vert2['coord'])],
        dtype=jigsaw_msh_t.VERT2_t
        )
    mesh.tria3 = np.array(
        [(index, id+1) for id, index in enumerate(mesh.tria3['index'])],
        dtype=jigsaw_msh_t.TRIA3_t
        )
    mesh.quad4 = np.array(
        [(index, id+1) for id, index in enumerate(mesh.quad4['index'])],
        dtype=jigsaw_msh_t.QUAD4_t
        )
    mesh.hexa8 = np.array(
        [(index, id+1) for id, index in enumerate(mesh.hexa8['index'])],
        dtype=jigsaw_msh_t.HEXA8_t
        )


def _get_sieve_mask(mesh, polygons, sieve_area):

    # NOTE: Some polygons are ghost polygons (interior)
    areas = [p.area for p in polygons.geoms]
    if sieve_area is None:
        remove = np.where(areas < np.max(areas))[0].tolist()
    else:
        remove = []
        for idx, patch_area in enumerate(areas):
            if patch_area <= sieve_area:
                remove.append(idx)

    # if the path surrounds the node, these need to be removed.
    vert2_mask = np.full((mesh.vert2['coord'].shape[0],), False)
    for idx in remove:
        path = Path(polygons.geoms[idx].exterior.coords, closed=True)
        vert2_mask = vert2_mask | path.contains_points(mesh.vert2['coord'])

    return vert2_mask


def _sieve_by_mask(mesh, sieve_mask):

    # if the path surrounds the node, these need to be removed.
    vert2_mask = sieve_mask.copy()

    # select any connected nodes; these ones are missed by
    # path.contains_point() because they are at the path edges.
    _idxs = np.where(vert2_mask)[0]
    conn_verts = get_surrounding_elem_verts(mesh, _idxs)
    vert2_mask[conn_verts] = True

    # Also, there might be some dangling triangles without neighbors,
    # which are also missed by path.contains_point()
    lone_elem_verts = get_lone_element_verts(mesh)
    vert2_mask[lone_elem_verts] = True

    # Mask out elements containing the unwanted nodes.
    tria3_mask = np.any(vert2_mask[mesh.tria3['index']], axis=1)

    # Tria and node removal and renumbering indexes ...
    tria3_id_tag = mesh.tria3['IDtag'].take(np.where(~tria3_mask)[0])
    tria3_index = mesh.tria3['index'][~tria3_mask, :].flatten()
    used_indexes = np.unique(tria3_index)
    node_indexes = np.arange(mesh.vert2['coord'].shape[0])
    renum = {old: new for new, old in enumerate(np.unique(tria3_index))}
    tria3_index = np.array([renum[i] for i in tria3_index])
    tria3_index = tria3_index.reshape((tria3_id_tag.shape[0], 3))
    vert2_idxs = np.where(np.isin(node_indexes, used_indexes))[0]

    # update vert2
    mesh.vert2 = mesh.vert2.take(vert2_idxs, axis=0)

    # update value
    if len(mesh.value) > 0:
        mesh.value = mesh.value.take(vert2_idxs, axis=0)

    # update tria3
    mesh.tria3 = np.array(
        [(tuple(indices), tria3_id_tag[i])
         for i, indices in enumerate(tria3_index)],
        dtype=jigsaw_msh_t.TRIA3_t)


def finalize_mesh(mesh, sieve_area=None):

    cleanup_isolates(mesh)

    while True:
        no_op = True

        pinched_nodes = get_pinched_nodes(mesh)
        if len(pinched_nodes):
            no_op = False
            # TODO drop fewer elements for pinch
            clip_mesh_by_vertex(
                mesh, pinched_nodes,
                can_use_other_verts=True,
                inverse=True, in_place=True)

        boundary_polys = get_mesh_polygons(mesh)
        sieve_mask = _get_sieve_mask(mesh, boundary_polys, sieve_area)
        if np.sum(sieve_mask):
            no_op = False
            _sieve_by_mask(mesh, sieve_mask)

        if no_op:
            break

    cleanup_isolates(mesh)
    put_id_tags(mesh)


def remesh_small_elements(opts, geom, mesh, hfun):

    """
    This function uses all the inputs for a given jigsaw meshing
    process and based on that finds and fixes tiny elements that
    might occur during initial meshing by iteratively remeshing
    """

    # TODO: Implement for quad, etc.


    hmin = np.min(hfun.value)
    equilat_area = np.sqrt(3)/4 * hmin**2
    # List of arbitrary coef of equilateral triangle area for a givven
    # minimum mesh size to come up with a decent cut off.
    coeffs = [0.5, 0.2, 0.1, 0.05]

    fixed_mesh = mesh
    for coef in coeffs:
        tria_areas = calculate_tria_areas(fixed_mesh)
        tiny_sz = coef * equilat_area
        tiny_verts = np.unique(fixed_mesh.tria3['index'][tria_areas < tiny_sz, :].ravel())
        if len(tiny_verts) == 0:
            break
        mesh_clip = clip_mesh_by_vertex(fixed_mesh, tiny_verts, inverse=True)

        fixed_mesh = jigsawpy.jigsaw_msh_t()
        fixed_mesh.mshID = 'euclidean-mesh'
        fixed_mesh.ndims = +2
        if hasattr(mesh, 'crs'):
            fixed_mesh.crs = mesh.crs

        jigsawpy.lib.jigsaw(
            opts, geom, fixed_mesh, init=mesh_clip, hfun=hfun)


    return fixed_mesh


def sieve(mesh, area=None):
    """
    A mesh can consist of multiple separate subdomins on as single structure.
    This functions removes subdomains which are equal or smaller than the
    provided area. Default behaviours is to remove all subdomains except the
    largest one.
    """
    # select the nodes to remove based on multipolygon areas
    multipolygon = geom_to_multipolygon(mesh)
    areas = [polygon.area for polygon in multipolygon]
    if area is None:
        remove = np.where(areas < np.max(areas))[0].tolist()
    else:
        remove = []
        for idx, patch_area in enumerate(areas):
            if patch_area <= area:
                remove.append(idx)

    # if the path surrounds the node, these need to be removed.
    vert2_mask = np.full((mesh.vert2['coord'].shape[0],), False)
    for idx in remove:
        path = Path(multipolygon[idx].exterior.coords, closed=True)
        vert2_mask = vert2_mask | path.contains_points(mesh.vert2['coord'])

    # select any connected nodes; these ones are missed by
    # path.contains_point() because they are at the path edges.
    _idxs = np.where(vert2_mask)[0]
    conn_verts = get_surrounding_elem_verts(mesh, _idxs)
    vert2_mask[conn_verts] = True

    # Also, there might be some dangling triangles without neighbors,
    # which are also missed by path.contains_point()
    lone_elem_verts = get_lone_element_verts(mesh)
    vert2_mask[lone_elem_verts] = True


    # Mask out elements containing the unwanted nodes.
    tria3_mask = np.any(vert2_mask[mesh.tria3['index']], axis=1)

    # Renumber indexes ...
    # isolated node removal does not require elimination of triangles from
    # the table, therefore the length of the indexes is constant.
    # We must simply renumber the tria3 indexes to match the new node indexes.
    # Essentially subtract one, but going from the bottom of the index table
    # to the top.
    used_indexes = np.unique(mesh.tria3['index'])
    node_indexes = np.arange(mesh.vert2['coord'].shape[0])
    tria3_idxs = np.where(~np.isin(node_indexes, used_indexes))[0]
    tria3_id_tag = mesh.tria3['IDtag'].take(np.where(~tria3_mask)[0])
    tria3_index = mesh.tria3['index'][~tria3_mask, :].flatten()
    for idx in reversed(tria3_idxs):
        tria3_index[np.where(tria3_index >= idx)] -= 1
    tria3_index = tria3_index.reshape((tria3_id_tag.shape[0], 3))
    vert2_idxs = np.where(np.isin(node_indexes, used_indexes))[0]

    # update vert2
    mesh.vert2 = mesh.vert2.take(vert2_idxs, axis=0)

    # update value
    if len(mesh.value) > 0:
        mesh.value = mesh.value.take(vert2_idxs, axis=0)

    # update tria3
    mesh.tria3 = np.array(
        [(tuple(indices), tria3_id_tag[i])
         for i, indices in enumerate(tria3_index)],
        dtype=jigsaw_msh_t.TRIA3_t)


def sort_edges(edges):

    if len(edges) == 0:
        return edges

    # start ordering the edges into linestrings
    edge_collection = []
    ordered_edges = [edges.pop(-1)]
    e0, e1 = [list(t) for t in zip(*edges)]
    while len(edges) > 0:

        if ordered_edges[-1][1] in e0:
            idx = e0.index(ordered_edges[-1][1])
            ordered_edges.append(edges.pop(idx))

        elif ordered_edges[0][0] in e1:
            idx = e1.index(ordered_edges[0][0])
            ordered_edges.insert(0, edges.pop(idx))

        elif ordered_edges[-1][1] in e1:
            idx = e1.index(ordered_edges[-1][1])
            ordered_edges.append(
                list(reversed(edges.pop(idx))))

        elif ordered_edges[0][0] in e0:
            idx = e0.index(ordered_edges[0][0])
            ordered_edges.insert(
                0, list(reversed(edges.pop(idx))))

        else:
            edge_collection.append(tuple(ordered_edges))
            idx = -1
            ordered_edges = [edges.pop(idx)]

        e0.pop(idx)
        e1.pop(idx)

    # finalize
    if len(edge_collection) == 0 and len(edges) == 0:
        edge_collection.append(tuple(ordered_edges))
    else:
        edge_collection.append(tuple(ordered_edges))

    return edge_collection


def index_ring_collection(mesh):

    # find boundary edges using triangulation neighbors table,
    # see: https://stackoverflow.com/a/23073229/7432462
    boundary_edges = []
    tri = mesh_to_tri(mesh)
    idxs = np.vstack(
        list(np.where(tri.neighbors == -1))).T
    for i, j in idxs:
        boundary_edges.append(
            (int(tri.triangles[i, j]),
                int(tri.triangles[i, (j+1) % 3])))
    init_idx_ring_coll = sort_edges(boundary_edges)
    # sort index_rings into corresponding "polygons"
    areas = []
    vertices = mesh.vert2['coord']
    for index_ring in init_idx_ring_coll:
        e0, _ = [list(t) for t in zip(*index_ring)]
        areas.append(float(Polygon(vertices[e0, :]).area))

    # maximum area must be main mesh
    idx = areas.index(np.max(areas))
    exterior = init_idx_ring_coll.pop(idx)
    areas.pop(idx)
    _id = 0
    idx_ring_coll = {}
    idx_ring_coll[_id] = {
        'exterior': np.asarray(exterior),
        'interiors': []
        }
    e0, e1 = [list(t) for t in zip(*exterior)]
    path = Path(vertices[e0 + [e0[0]], :], closed=True)
    while len(init_idx_ring_coll) > 0:
        # find all internal rings
        potential_interiors = []
        for i, index_ring in enumerate(init_idx_ring_coll):
            e0, e1 = [list(t) for t in zip(*index_ring)]
            if path.contains_point(vertices[e0[0], :]):
                potential_interiors.append(i)
        # filter out nested rings
        real_interiors = []
        for i, p_interior in reversed(list(enumerate(potential_interiors))):
            _p_interior = init_idx_ring_coll[p_interior]
            check = [init_idx_ring_coll[_]
                     for j, _ in reversed(list(enumerate(potential_interiors)))
                     if i != j]
            has_parent = False
            for _path in check:
                e0, e1 = [list(t) for t in zip(*_path)]
                _path = Path(vertices[e0 + [e0[0]], :], closed=True)
                if _path.contains_point(vertices[_p_interior[0][0], :]):
                    has_parent = True
                    break
            if not has_parent:
                real_interiors.append(p_interior)
        # pop real rings from collection
        for i in reversed(sorted(real_interiors)):
            idx_ring_coll[_id]['interiors'].append(
                np.asarray(init_idx_ring_coll.pop(i)))
            areas.pop(i)
        # if no internal rings found, initialize next polygon
        if len(init_idx_ring_coll) > 0:
            idx = areas.index(np.max(areas))
            exterior = init_idx_ring_coll.pop(idx)
            areas.pop(idx)
            _id += 1
            idx_ring_coll[_id] = {
                'exterior': np.asarray(exterior),
                'interiors': []
                }
            e0, e1 = [list(t) for t in zip(*exterior)]
            path = Path(vertices[e0 + [e0[0]], :], closed=True)
    return idx_ring_coll


def outer_ring_collection(mesh):
    idx_ring_coll = index_ring_collection(mesh)
    exterior_ring_collection = defaultdict()
    for key, ring in idx_ring_coll.items():
        exterior_ring_collection[key] = ring['exterior']
    return exterior_ring_collection


def inner_ring_collection(mesh):
    idx_ring_coll = index_ring_collection(mesh)
    inner_ring_coll = defaultdict()
    for key, rings in idx_ring_coll.items():
        inner_ring_coll[key] = rings['interiors']
    return inner_ring_coll


def get_multipolygon_from_pathplot(ax):
    # extract linear_rings from plot
    linear_ring_collection = []
    for path_collection in ax.collections:
        for path in path_collection.get_paths():
            polygons = path.to_polygons(closed_only=True)
            for linear_ring in polygons:
                if linear_ring.shape[0] > 3:
                    linear_ring_collection.append(
                        LinearRing(linear_ring))
    if len(linear_ring_collection) > 1:
        # reorder linear rings from above
        polygon_collection = []
        while len(linear_ring_collection) > 0:
            areas = [Polygon(linear_ring).area
                     for linear_ring in linear_ring_collection]
            idx = np.where(areas == np.max(areas))[0][0]
            outer_ring = linear_ring_collection.pop(idx)
            path = Path(np.asarray(outer_ring.coords), closed=True)

            inner_rings = []
            for i, linear_ring in reversed(
                    list(enumerate(linear_ring_collection))):
                xy = np.asarray(linear_ring.coords)[0, :]
                if path.contains_point(xy):
                    inner_rings.append(linear_ring_collection.pop(i))
            polygon_collection.append(Polygon(outer_ring, inner_rings))

        multipolygon = MultiPolygon(polygon_collection)
    else:
        multipolygon = MultiPolygon(
            [Polygon(linear_ring_collection.pop())])
    return multipolygon


def signed_polygon_area(vertices):
    # https://code.activestate.com/recipes/578047-area-of-polygon-using-shoelace-formula/
    n = len(vertices)  # of vertices
    area = 0.0
    for i in range(n):
        j = (i + 1) % n
        area += vertices[i][0] * vertices[j][1]
        area -= vertices[j][0] * vertices[i][1]
    return area / 2.0


def vertices_around_vertex(mesh):
    if mesh.mshID == 'euclidean-mesh':
        def append(geom):
            for simplex in geom['index']:
                for i, j in permutations(simplex, 2):
                    vert_list[i].add(j)
        vert_list = defaultdict(set)
        append(mesh.tria3)
        append(mesh.quad4)
        append(mesh.hexa8)
        return vert_list

    msg = f"Not implemented for mshID={mesh.mshID}"
    raise NotImplementedError(msg)

def get_surrounding_elem_verts(mesh, in_vert):

    '''
    Find vertices of elements connected to input vertices
    '''

    tria = mesh.tria3['index']
    quad = mesh.quad4['index']
    hexa = mesh.hexa8['index']

    # NOTE: np.any is used so that vertices that are not in in_verts
    # triangles but are part of the triangles that include in_verts
    # are considered too
    mark_tria = np.any(
            (np.isin(tria.ravel(), in_vert).reshape(
                tria.shape)), 1)
    mark_quad = np.any(
            (np.isin(quad.ravel(), in_vert).reshape(
                quad.shape)), 1)
    mark_hexa = np.any(
            (np.isin(hexa.ravel(), in_vert).reshape(
                hexa.shape)), 1)

    conn_verts = np.unique(np.concatenate(
        (tria[mark_tria, :].ravel(),
         quad[mark_quad, :].ravel(),
         hexa[mark_hexa, :].ravel())))

    return conn_verts

def get_lone_element_verts(mesh):

    '''
    Also, there might be some dangling triangles without neighbors,
    which are also missed by path.contains_point()
    '''

    tria = mesh.tria3['index']
    quad = mesh.quad4['index']
    hexa = mesh.hexa8['index']

    unq_verts, counts = np.unique(
        np.concatenate((tria.ravel(), quad.ravel(), hexa.ravel())),
        return_counts=True)
    once_verts = unq_verts[counts == 1]

    # NOTE: np.all so that lone elements are found vs elements that
    # have nodes that are used only once
    mark_tria = np.all(
            (np.isin(tria.ravel(), once_verts).reshape(
                tria.shape)), 1)
    mark_quad = np.all(
            (np.isin(quad.ravel(), once_verts).reshape(
                quad.shape)), 1)
    mark_hexa = np.all(
            (np.isin(hexa.ravel(), once_verts).reshape(
                hexa.shape)), 1)

    lone_elem_verts = np.unique(np.concatenate(
        (tria[mark_tria, :].ravel(),
         quad[mark_quad, :].ravel(),
         hexa[mark_hexa, :].ravel())))

    return lone_elem_verts



# https://en.wikipedia.org/wiki/Polygon_mesh#Summary_of_mesh_representation
# V-V     All vertices around vertex
# E-F     All edges of a face
# V-F     All vertices of a face
# F-V     All faces around a vertex
# E-V     All edges around a vertex
# F-E     Both faces of an edge
# V-E     Both vertices of an edge
# Flook   Find face with given vertices
def get_verts_in_shape(
        mesh: jigsaw_msh_t,
        shape: Union[box, Polygon, MultiPolygon],
        from_box: bool = False,
        num_adjacent: int = 0
        ) -> Sequence[int]:

    if from_box:
        crd = mesh.vert2['coord']

        xmin, ymin, xmax, ymax = shape.bounds

        in_box_idx_1 = np.arange(len(crd))[crd[:, 0] > xmin]
        in_box_idx_2 = np.arange(len(crd))[crd[:, 0] < xmax]
        in_box_idx_3 = np.arange(len(crd))[crd[:, 1] > ymin]
        in_box_idx_4 = np.arange(len(crd))[crd[:, 1] < ymax]
        in_box_idx = reduce(
            np.intersect1d, (in_box_idx_1, in_box_idx_2,
                             in_box_idx_3, in_box_idx_4))
        return in_box_idx


    pt_series = gpd.GeoSeries(gpd.points_from_xy(
        mesh.vert2['coord'][:,0], mesh.vert2['coord'][:,1]))
    shp_series = gpd.GeoSeries(shape)

    in_shp_idx = pt_series.sindex.query_bulk(
            shp_series, predicate="intersects")

    in_shp_idx = select_adjacent(mesh, in_shp_idx, num_layers=num_adjacent)

    return in_shp_idx

def select_adjacent(mesh, in_indices, num_layers):

    selected_indices = in_indices.copy()

    if mesh.mshID == 'euclidean-mesh' and mesh.ndims == 2:

        for i in range(num_layers - 1):

            coord = mesh.vert2['coord']

            # TODO: What about edge2
            mesh_types = {
                'tria3': 'TRIA3_t',
                'quad4': 'QUAD4_t',
                'hexa8': 'HEXA8_t'
            }
            elm_dict = {
                key: getattr(mesh, key)['index'] for key in mesh_types}

            mark_func = np.any

            mark_dict = {
                key: mark_func(
                    (np.isin(elems.ravel(), selected_indices).reshape(
                        elems.shape)), 1)
                for key, elems in elm_dict.items()}

            picked_elems = {
                    key: elm_dict[key][mark_dict[key], :]
                    for key in elm_dict}

            selected_indices = np.unique(np.concatenate(
                [pick.ravel() for pick in picked_elems.values()]))

        return selected_indices


    msg = (f"Not implemented for"
           f" mshID={mesh.mshID} and dim={mesh.ndims}")
    raise NotImplementedError(msg)


@must_be_euclidean_mesh
def get_cross_edges(
        mesh: jigsaw_msh_t,
        shape: Union[box, Polygon, MultiPolygon],
        ) -> Sequence[Tuple[int, int]]:

    '''
    Return the list of edges crossing the input shape exterior
    '''

    coords = mesh.vert2['coord']

    coord_dict = {}
    for i, coo in enumerate(coords):
        coord_dict[tuple(coo)] = i

    gdf_shape = gpd.GeoDataFrame(
            geometry=gpd.GeoSeries(shape))
    exteriors = [pl.exterior for pl in gdf_shape.explode(index_parts=True).geometry]

    # TODO: Reduce domain of search for faster results
    all_edges = get_mesh_edges(mesh, unique=True)
    edge_coords = coords[all_edges, :]
    gdf_edg = gpd.GeoDataFrame(
        geometry=gpd.GeoSeries(linemerge(edge_coords.tolist())))

    gdf_x = gpd.sjoin(
            gdf_edg.explode(index_parts=True),
            gpd.GeoDataFrame(geometry=gpd.GeoSeries(exteriors)),
            how='inner', predicate='intersects')


    cut_coords = [
        list(cooseq)
        for cooseq in gdf_x.geometry.apply(lambda i: i.coords).values]

    cut_edges = np.array([
        (coo_list[i], coo_list[i+1])
        for coo_list in cut_coords
        for i in range(len(coo_list)-1) ])

    cut_edge_idx = np.array(
            [coord_dict[tuple(coo)]
             for coo in cut_edges.reshape(-1, 2)]).reshape(
                     cut_edges.shape[:2])

    return cut_edge_idx



def clip_mesh_by_shape(
        mesh: jigsaw_msh_t,
        shape: Union[box, Polygon, MultiPolygon],
        use_box_only: bool = False,
        fit_inside: bool = True,
        inverse: bool = False,
        in_place: bool = False,
        check_cross_edges: bool = False,
        adjacent_layers: int = 0
        ) -> jigsaw_msh_t:


    # NOTE: Checking cross edge is only meaningful when
    # fit inside flag is NOT set
    edge_flag = check_cross_edges and not fit_inside

    # If we want to calculate inverse based on shape, calculating
    # from bbox first results in the wrong result
    if not inverse or use_box_only:

        # First based on bounding box only
        shape_box = box(*shape.bounds)

        # TODO: Optimize for multipolygons (use separate bboxes)
        in_box_idx = get_verts_in_shape(mesh, shape_box, True, adjacent_layers)

        if edge_flag and not inverse:
            x_edge_idx = get_cross_edges(mesh, shape_box)
            in_box_idx = np.append(in_box_idx, np.unique(x_edge_idx))

        mesh = clip_mesh_by_vertex(
                mesh, in_box_idx, not fit_inside, inverse, in_place)

        if use_box_only:
            if edge_flag and inverse:
                x_edge_idx = get_cross_edges(mesh, shape_box)
                mesh = remove_mesh_by_edge(
                        mesh, x_edge_idx, in_place)
            return mesh

    in_shp_idx = get_verts_in_shape(mesh, shape, False, adjacent_layers)

    if edge_flag and not inverse:
        x_edge_idx = get_cross_edges(mesh, shape)
        in_shp_idx = np.append(in_shp_idx, np.unique(x_edge_idx))

    mesh = clip_mesh_by_vertex(
            mesh, in_shp_idx, not fit_inside, inverse, in_place)

    if edge_flag and inverse:
        x_edge_idx = get_cross_edges(mesh, shape)
        mesh = remove_mesh_by_edge(mesh, x_edge_idx, in_place)

    return mesh


def remove_mesh_by_edge(
        mesh: jigsaw_msh_t,
        edges: Sequence[Tuple[int, int]],
        in_place: bool = False
        ) -> jigsaw_msh_t:

    mesh_out = mesh
    if not in_place:
        mesh_out = deepcopy(mesh)

    # NOTE: This method selects more elements than needed as it
    # uses only existance of more than two of the vertices attached
    # to the input edges in the element as criteria.
    edge_verts = np.unique(edges)

    for etype in ELEM_2D_TYPES:
        elems = getattr(mesh, etype)['index']
        # If a given element contains to vertices from
        # a crossing edge, it is selected
        test = np.sum(np.isin(elems, edge_verts), axis=1)
        elems = elems[test < 2]
        setattr(mesh_out, etype, np.array(
                [(idx, 0) for idx in elems],
                dtype=getattr(
                    jigsawpy.jigsaw_msh_t, f'{etype.upper()}_t')))

    return mesh_out


def clip_mesh_by_vertex(
        mesh: jigsaw_msh_t,
        vert_in: Sequence[int],
        can_use_other_verts: bool = False,
        inverse: bool = False,
        in_place: bool = False
        ) -> jigsaw_msh_t:


    if mesh.mshID == 'euclidean-mesh' and mesh.ndims == 2:
        coord = mesh.vert2['coord']

        # TODO: What about edge2 if in_place?
        mesh_types = {
            'tria3': 'TRIA3_t',
            'quad4': 'QUAD4_t',
            'hexa8': 'HEXA8_t'
        }
        elm_dict = {
            key: getattr(mesh, key)['index'] for key in mesh_types}

        # Whether elements that include "in"-vertices can be created
        # using vertices other than "in"-vertices
        mark_func = np.all
        if can_use_other_verts:
            mark_func = np.any

        mark_dict = {
            key: mark_func(
                (np.isin(elems.ravel(), vert_in).reshape(
                    elems.shape)), 1)
            for key, elems in elm_dict.items()}


        # Whether to return elements found by "in" vertices or return
        # all elements except them
        if inverse:
            mark_dict = {
                key: np.logical_not(mark)
                for key, mark in mark_dict.items()}

        # Find elements based on old vertex index
        elem_draft_dict = {
                key: elm_dict[key][mark_dict[key], :]
                for key in elm_dict}

        crd_old_to_new = {
                index: i for i, index
                in enumerate(sorted(np.unique(np.concatenate(
                        [draft.ravel()
                            for draft in elem_draft_dict.values()]
                        ))))
            }

        elem_final_dict = {
            key: np.array(
                [[crd_old_to_new[x] for x in  element]
                 for element in draft])
            for key, draft in elem_draft_dict.items()
        }

        new_coord = coord[list(crd_old_to_new.keys()), :]
        value = np.zeros(shape=(0, 0), dtype=jigsaw_msh_t.REALS_t)
        if len(mesh.value) == len(coord):
            value = mesh.value.take(
                    list(crd_old_to_new.keys()), axis=0).copy()


        mesh_out = mesh
        if not in_place:
            mesh_out = jigsaw_msh_t()
            mesh_out.mshID = mesh.mshID
            mesh_out.ndims = mesh.ndims
            if hasattr(mesh, "crs"):
                mesh_out.crs = deepcopy(mesh.crs)

        mesh_out.value = value

        mesh_out.vert2 = np.array(
            [(coo, 0) for coo in new_coord],
            dtype=jigsaw_msh_t.VERT2_t)

        for key, elem_type in mesh_types.items():
            setattr(
                mesh_out,
                key,
                np.array(
                    [(con, 0) for con in elem_final_dict[key]],
                    dtype=getattr(jigsaw_msh_t, elem_type)))

        return mesh_out

    msg = (f"Not implemented for"
           f" mshID={mesh.mshID} and dim={mesh.ndims}")
    raise NotImplementedError(msg)






@must_be_euclidean_mesh
def get_mesh_edges(mesh: jigsaw_msh_t, unique=True):

    # NOTE: For msh_t type vertex id and index are the same
    trias = mesh.tria3['index']
    quads = mesh.quad4['index']
    hexas = mesh.hexa8['index']

    # Get unique set of edges by rolling connectivity
    # and joining connectivities in 3rd dimension, then sorting
    # to get all edges with lower index first
    all_edges = np.empty(shape=(0, 2), dtype=trias.dtype)
    for elm_type in [trias, quads, hexas]:
        if elm_type.shape[0]:
            edges = np.sort(
                    np.stack(
                        (elm_type, np.roll(elm_type, shift=1, axis=1)),
                        axis=2),
                    axis=2)
            edges = edges.reshape(np.product(edges.shape[0:2]), 2)
            all_edges = np.vstack((all_edges, edges))

    if unique:
        all_edges = np.unique(all_edges, axis=0)

    return all_edges


@must_be_euclidean_mesh
def calculate_tria_areas(mesh):

    coord = mesh.vert2['coord']
    trias = mesh.tria3['index']

    tria_coo = coord[
        np.sort(np.stack((trias, np.roll(trias, shift=1, axis=1)),
                         axis=2),
                axis=2)]
    tria_side_components = np.diff(tria_coo, axis=2).squeeze()
    tria_sides = np.sqrt(
            np.sum(np.power(np.abs(tria_side_components), 2),
                   axis=2).squeeze())
    perimeter = np.sum(tria_sides, axis=1) / 2
    perimeter = perimeter.reshape(len(perimeter), 1)
    # pylint: disable=W0632
    a_side, b_side, c_side = np.split(tria_sides, 3, axis=1)
    tria_areas = np.sqrt(
            perimeter*(perimeter-a_side)
            * (perimeter-b_side)*(perimeter-c_side)
            ).squeeze()
    return tria_areas

@must_be_euclidean_mesh
def calculate_edge_lengths(mesh, transformer=None):

    coord = mesh.vert2['coord']
    if transformer is not None:
        coord = np.vstack(
            transformer.transform(coord[:, 0], coord[:, 1])).T

    # Get unique set of edges by rolling connectivity
    # and joining connectivities in 3rd dimension, then sorting
    # to get all edges with lower index first
    all_edges = get_mesh_edges(mesh, unique=True)

    # ONLY TESTED FOR TRIA AS OF NOW

    # This part of the function is generic for tria and quad

    # Get coordinates for all edge vertices
    edge_coords = coord[all_edges, :]

    # Calculate length of all edges based on acquired coords
    edge_lens = np.sqrt(
            np.sum(
                np.power(
                    np.abs(np.diff(edge_coords, axis=1)), 2)
                ,axis=2)).squeeze()

    edge_dict = defaultdict(float)
    for i, edge in enumerate(all_edges):
        edge_dict[tuple(edge)] = edge_lens[i]

    return edge_dict


@must_be_euclidean_mesh
def elements(mesh):
    elements_id = []
    elements_id.extend(list(mesh.tria3['IDtag']))
    elements_id.extend(list(mesh.quad4['IDtag']))
    elements_id.extend(list(mesh.hexa8['IDtag']))
    elements_id = range(1, len(elements_id)+1) \
        if len(set(elements_id)) != len(elements_id) else elements_id
    elems = []
    elems.extend(list(mesh.tria3['index']))
    elems.extend(list(mesh.quad4['index']))
    elems.extend(list(mesh.hexa8['index']))
    elems = {
        elements_id[i]: indexes for i, indexes in enumerate(elems)}
    return elems


@must_be_euclidean_mesh
def faces_around_vertex(mesh):
    _elements = elements(mesh)
    length = max(map(len, _elements.values()))
    y = np.array([xi+[-99999]*(length-len(xi)) for xi in _elements.values()])
    faces_around_vert = defaultdict(set)
    for i, coord in enumerate(mesh.vert2['index']):
        # TODO:
        pass
#        np.isin(i, axis=0)
#        faces_around_vert[i].add()

    faces_around_vert = defaultdict(set)


def get_boundary_edges(mesh):

    '''
    Find internal and external boundaries of mesh
    '''

    coord = mesh.vert2['coord']

    all_edges = get_mesh_edges(mesh, unique=False)

    # Simplexes (list of node indices)
    all_edges, e_cnt = np.unique(all_edges, axis=0, return_counts=True)
    shared_edges = all_edges[e_cnt == 2]
    boundary_edges = all_edges[e_cnt == 1]

    return boundary_edges


def get_pinched_nodes(mesh):

    '''
    Find nodes through which fluid cannot flow
    '''

    boundary_edges = get_boundary_edges(mesh)

    # Node indices
    boundary_verts, vb_cnt = np.unique(boundary_edges, return_counts=True)

    # vertices/nodes that have more than 2 boundary edges are pinch
    pinch_verts = boundary_verts[vb_cnt > 2]

    return pinch_verts


def has_pinched_nodes(mesh):

    # Older function: computationally more expensive and missing some
    # nodes

    _inner_ring_collection = inner_ring_collection(mesh)
    all_nodes = []
    for inner_rings in _inner_ring_collection.values():
        for ring in inner_rings:
            all_nodes.extend(np.asarray(ring)[:, 0].tolist())
    u, c = np.unique(all_nodes, return_counts=True)
    if len(u[c > 1]) > 0:
        return True

    return False


def cleanup_pinched_nodes(mesh):

    # Older function: computationally more expensive and missing some
    # nodes

    _inner_ring_collection = inner_ring_collection(mesh)
    all_nodes = []
    for inner_rings in _inner_ring_collection.values():
        for ring in inner_rings:
            all_nodes.extend(np.asarray(ring)[:, 0].tolist())
    u, c = np.unique(all_nodes, return_counts=True)
    mesh.tria3 = mesh.tria3.take(
        np.where(
            ~np.any(np.isin(mesh.tria3['index'], u[c > 1]), axis=1))[0],
        axis=0)


def interpolate(src: jigsaw_msh_t, dst: jigsaw_msh_t, **kwargs):
    if src.mshID == 'euclidean-grid' and dst.mshID == 'euclidean-mesh':
        interpolate_euclidean_grid_to_euclidean_mesh(src, dst, **kwargs)
    elif src.mshID == 'euclidean-mesh' and dst.mshID == 'euclidean-mesh':
        interpolate_euclidean_mesh_to_euclidean_mesh(src, dst, **kwargs)
    else:
        raise NotImplementedError(
            f'Not implemented type combination: source={src.mshID}, '
            f'dest={dst.mshID}')


def interpolate_euclidean_mesh_to_euclidean_mesh(
        src: jigsaw_msh_t,
        dst: jigsaw_msh_t,
        method='linear',
        fill_value=np.nan
):
    values = griddata(
        src.vert2['coord'],
        src.value.flatten(),
        dst.vert2['coord'],
        method=method,
        fill_value=fill_value
    )
    dst.value = np.array(
        values.reshape(len(values), 1), dtype=jigsaw_msh_t.REALS_t)


def interpolate_euclidean_grid_to_euclidean_mesh(
        src: jigsaw_msh_t,
        dst: jigsaw_msh_t,
        bbox=None,
        kx=3,
        ky=3,
        s=0
):
    values = RectBivariateSpline(
        src.xgrid,
        src.ygrid,
        src.value.T,
        bbox=bbox or [None, None, None, None],
        kx=kx,
        ky=ky,
        s=s
        ).ev(
        dst.vert2['coord'][:, 0],
        dst.vert2['coord'][:, 1])
    dst.value = np.array(
        values.reshape((values.size, 1)),
        dtype=jigsaw_msh_t.REALS_t)


def tricontourf(
    mesh,
    ax=None,
    show=False,
    figsize=None,
    extend='both',
    colorbar=False,
    **kwargs
):
    if ax is None:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)
    tcf = ax.tricontourf(
        mesh.vert2['coord'][:, 0],
        mesh.vert2['coord'][:, 1],
        mesh.tria3['index'],
        mesh.value.flatten(),
        **kwargs)
    if colorbar:
        plt.colorbar(tcf)
    if show:
        plt.gca().axis('scaled')
        plt.show()
    return ax


def triplot(
    mesh,
    axes=None,
    show=False,
    figsize=None,
    color='k',
    linewidth=0.07,
    **kwargs
):
    if axes is None:
        fig = plt.figure(figsize=figsize)
        axes = fig.add_subplot(111)
    axes.triplot(
        mesh.vert2['coord'][:, 0],
        mesh.vert2['coord'][:, 1],
        mesh.tria3['index'],
        color=color,
        linewidth=linewidth,
        **kwargs)
    if show:
        axes.axis('scaled')
        plt.show()
    return axes


def reproject(
        mesh: jigsaw_msh_t,
        dst_crs: Union[str, CRS]
):
    src_crs = mesh.crs
    dst_crs = CRS.from_user_input(dst_crs)
    transformer = Transformer.from_crs(src_crs, dst_crs, always_xy=True)
    # pylint: disable=E0633
    x, y = transformer.transform(
        mesh.vert2['coord'][:, 0], mesh.vert2['coord'][:, 1])
    mesh.vert2 = np.array(
        [([x[i], y[i]], mesh.vert2['IDtag'][i]) for i
         in range(len(mesh.vert2['IDtag']))],
        dtype=jigsaw_msh_t.VERT2_t)
    mesh.crs = dst_crs


def limgrad(mesh, dfdx, imax=100):
    """
    See https://github.com/dengwirda/mesh2d/blob/master/hjac-util/limgrad.m
    for original source code.
    """
    tri = mesh_to_tri(mesh)
    xy = np.vstack([tri.x, tri.y]).T
    edge = tri.edges
    dx = np.subtract(xy[edge[:, 0], 0], xy[edge[:, 1], 0])
    dy = np.subtract(xy[edge[:, 0], 1], xy[edge[:, 1], 1])
    elen = np.sqrt(dx**2+dy**2)
    ffun = mesh.value.flatten()
    aset = np.zeros(ffun.shape)
    ftol = np.min(ffun) * np.sqrt(np.finfo(float).eps)
    # precompute neighbor table
    point_neighbors = defaultdict(set)
    for simplex in tri.triangles:
        for i, j in permutations(simplex, 2):
            point_neighbors[i].add(j)
    # iterative smoothing
    for _iter in range(1, imax+1):
        aidx = np.where(aset == _iter-1)[0]
        if len(aidx) == 0.:
            break
        active_idxs = np.argsort(ffun[aidx])
        for active_idx in active_idxs:
            adjacent_edges = point_neighbors[active_idx]
            for adj_edge in adjacent_edges:
                if ffun[adj_edge] > ffun[active_idx]:
                    fun1 = ffun[active_idx] + elen[active_idx] * dfdx
                    if ffun[adj_edge] > fun1+ftol:
                        ffun[adj_edge] = fun1
                        aset[adj_edge] = _iter
                else:
                    fun2 = ffun[adj_edge] + elen[active_idx] * dfdx
                    if ffun[active_idx] > fun2+ftol:
                        ffun[active_idx] = fun2
                        aset[active_idx] = _iter
    if not _iter < imax:
        msg = f'limgrad() did not converge within {imax} iterations.'
        raise Exception(msg)
    return ffun


def msh_t_to_grd(msh: jigsaw_msh_t) -> Dict:

    src_crs = msh.crs if hasattr(msh, 'crs') else None
    coords = msh.vert2['coord']
    desc = "EPSG:4326"
    if src_crs is not None:
        # TODO: Support non EPSG:4326 CRS
#        desc = src_crs.to_string()
        epsg_4326 = CRS.from_epsg(4326)
        if not src_crs.equals(epsg_4326):
            transformer = Transformer.from_crs(
                src_crs, epsg_4326, always_xy=True)
            coords = np.vstack(
                transformer.transform(coords[:, 0], coords[:, 1])).T

    nodes = {
        i + 1: [tuple(p.tolist()), v] for i, (p, v) in
            enumerate(zip(coords, -msh.value))}
    # NOTE: Node IDs are node index + 1
    elems = {
        i + 1: v + 1 for i, v in enumerate(msh.tria3['index'])}
    offset = len(elems)
    elems.update({
        offset + i + 1: v + 1 for i, v in enumerate(msh.quad4['index'])})

    return {'description': desc,
            'nodes': nodes,
            'elements': elems}


def grd_to_msh_t(_grd: Dict) -> jigsaw_msh_t:

    msh = jigsaw_msh_t()
    msh.ndims = +2
    msh.mshID = 'euclidean-mesh'
    id_to_index = {node_id: index for index, node_id
                   in enumerate(_grd['nodes'].keys())}
    triangles = [list(map(lambda x: id_to_index[x], element)) for element
                 in _grd['elements'].values() if len(element) == 3]
    quads = [list(map(lambda x: id_to_index[x], element)) for element
             in _grd['elements'].values() if len(element) == 4]
    msh.vert2 = np.array([(coord, 0) for coord, _ in _grd['nodes'].values()],
                         dtype=jigsaw_msh_t.VERT2_t)
    msh.tria3 = np.array([(index, 0) for index in triangles],
                         dtype=jigsaw_msh_t.TRIA3_t)
    msh.quad4 = np.array([(index, 0) for index in quads],
                         dtype=jigsaw_msh_t.QUAD4_t)
    value = [value for _, value in _grd['nodes'].values()]
    msh.value = np.array(np.array(value).reshape((len(value), 1)),
                         dtype=jigsaw_msh_t.REALS_t)
    crs = _grd.get('crs')
    if crs is not None:
        msh.crs = CRS.from_user_input(crs)
    return msh


def msh_t_to_2dm(msh: jigsaw_msh_t):
    coords = msh.vert2['coord']
    src_crs = msh.crs if hasattr(msh, 'crs') else None
    if src_crs is not None:
        epsg_4326 = CRS.from_epsg(4326)
        if not src_crs.equals(epsg_4326):
            transformer = Transformer.from_crs(
                src_crs, epsg_4326, always_xy=True)
            coords = np.vstack(
                transformer.transform(coords[:, 0], coords[:, 1])).T
    return {
            'ND': {i+1: (coord, msh.value[i, 0] if not
                         np.isnan(msh.value[i, 0]) else -99999)
                   for i, coord in enumerate(coords)},
            'E3T': {i+1: index+1 for i, index
                    in enumerate(msh.tria3['index'])},
            'E4Q': {i+1: index+1 for i, index
                    in enumerate(msh.quad4['index'])}
        }


def sms2dm_to_msh_t(_sms2dm: Dict) -> jigsaw_msh_t:
    msh = jigsaw_msh_t()
    msh.ndims = +2
    msh.mshID = 'euclidean-mesh'
    id_to_index = {node_id: index for index, node_id
                   in enumerate(_sms2dm['ND'].keys())}
    if 'E3T' in _sms2dm:
        triangles = [list(map(lambda x: id_to_index[x], element)) for element
                     in _sms2dm['E3T'].values()]
        msh.tria3 = np.array([(index, 0) for index in triangles],
                             dtype=jigsaw_msh_t.TRIA3_t)
    if 'E4Q' in _sms2dm:
        quads = [list(map(lambda x: id_to_index[x], element)) for element
                 in _sms2dm['E4Q'].values()]
        msh.quad4 = np.array([(index, 0) for index in quads],
                             dtype=jigsaw_msh_t.QUAD4_t)
    msh.vert2 = np.array([(coord, 0) for coord, _ in _sms2dm['ND'].values()],
                         dtype=jigsaw_msh_t.VERT2_t)
    value = [value for _, value in _sms2dm['ND'].values()]
    msh.value = np.array(np.array(value).reshape((len(value), 1)),
                         dtype=jigsaw_msh_t.REALS_t)
    crs = _sms2dm.get('crs')
    if crs is not None:
        msh.crs = CRS.from_user_input(crs)
    return msh

@must_be_euclidean_mesh
def msh_t_to_utm(msh):
    utm_crs = estimate_mesh_utm(msh)
    if utm_crs is None:
        return

    transformer = Transformer.from_crs(
        msh.crs, utm_crs, always_xy=True)

    coords = msh.vert2['coord']

    # pylint: disable=E0633
    coords[:, 0], coords[:, 1] = transformer.transform(
            coords[:, 0], coords[:, 1])
    msh.vert2['coord'][:] = coords
    msh.crs = utm_crs


def estimate_bounds_utm(bounds, crs="EPSG:4326"):
    in_crs = CRS.from_user_input(crs)
    if in_crs.is_geographic:
        x0, y0, x1, y1 = bounds
        _, _, number, letter = utm.from_latlon(
                (y0 + y1)/2, (x0 + x1)/2)
        # PyProj 3.2.1 throws error if letter is provided
        utm_crs = CRS(
                proj='utm',
                zone=f'{number}',
                south=(y0 + y1)/2 < 0,
                ellps={
                    'GRS 1980': 'GRS80',
                    'WGS 84': 'WGS84'
                    }[in_crs.ellipsoid.name]
            )
        return utm_crs

    return None

@must_be_euclidean_mesh
def estimate_mesh_utm(msh):
    if hasattr(msh, 'crs'):
        coords = msh.vert2['coord']
        x0, y0, x1, y1 = (
            np.min(coords[:, 0]), np.min(coords[:, 1]),
            np.max(coords[:, 0]), np.max(coords[:, 1]))
        utm_crs = estimate_bounds_utm((x0, y0, x1, y1), msh.crs)
        return utm_crs

    return None

def get_polygon_channels(polygon, width, simplify=None, join_style=3):

    # Operations are done without any CRS info consideration

    polys_gdf = gpd.GeoDataFrame(
        geometry=gpd.GeoSeries(polygon))

    if isinstance(simplify, (int, float)):
        polys_gdf = gpd.GeoDataFrame(
                geometry=polys_gdf.simplify(
                    tolerance=simplify,
                    preserve_topology=False))

    buffer_size = width/2
    buffered_gdf = gpd.GeoDataFrame(
            geometry=polys_gdf.buffer(-buffer_size).buffer(
                    buffer_size,
                    join_style=join_style))

    buffered_gdf = buffered_gdf[~buffered_gdf.is_empty]
    if len(buffered_gdf) == 0:
        # All is channel!
        return polygon
    channels_gdf = gpd.overlay(
            polys_gdf, buffered_gdf, how='difference')

    # Use square - 1/4 circle as cleanup criteria
    channels_gdf = gpd.GeoDataFrame(
           geometry=gpd.GeoSeries(
               [p for i in channels_gdf.geometry
                   for p in i.geoms
                   if p.area > width**2 * (1-np.pi/4)]))


    ret_val = channels_gdf.unary_union
    if isinstance(ret_val, GeometryCollection):
        return None

    if isinstance(ret_val, Polygon):
        ret_val = MultiPolygon([ret_val])

    return ret_val


def merge_msh_t(
        *mesh_list,
        out_crs="EPSG:4326",
        drop_by_bbox=True,
        can_overlap=True,
        check_cross_edges=False):


    dst_crs = CRS.from_user_input(out_crs)

    mesh_types = {
        'tria3': 'TRIA3_t',
        'quad4': 'QUAD4_t',
        'hexa8': 'HEXA8_t'
    }

    coord = []
    elems = {k: [] for k in mesh_types}
    value = []
    offset = 0

    mesh_shape_list = []
    # Last has the highest priority
    for mesh in mesh_list[::-1]:
        if not dst_crs.equals(mesh.crs):
            # To avoid modifying inputs
            mesh = deepcopy(mesh)
            reproject(mesh, dst_crs)

        if drop_by_bbox:
            x = mesh.vert2['coord'][:, 0]
            y = mesh.vert2['coord'][:, 1]
            mesh_shape = box(np.min(x), np.min(y), np.max(x), np.max(y))
        else:
            mesh_shape = get_mesh_polygons(mesh)

        for ishp in mesh_shape_list:
            # NOTE: fit_inside = True w/ inverse = True results
            # in overlap when clipping low-priority mesh
            mesh = clip_mesh_by_shape(
                mesh, ishp,
                use_box_only=drop_by_bbox,
                fit_inside=can_overlap,
                inverse=True,
                check_cross_edges=check_cross_edges)

        mesh_shape_list.append(mesh_shape)


        for k in mesh_types:
            cnn = getattr(mesh, k)
            elems[k].append(cnn['index'] + offset)
        coord.append(mesh.vert2['coord'])
        value.append(mesh.value)
        offset += coord[-1].shape[0]

    composite_mesh = jigsaw_msh_t()
    composite_mesh.mshID = 'euclidean-mesh'
    composite_mesh.ndims = 2

    composite_mesh.vert2 = np.array(
            [(coord, 0) for coord in np.vstack(coord)],
            dtype=jigsaw_msh_t.VERT2_t)
    composite_mesh.value = np.array(
            np.vstack(value),
            dtype=jigsaw_msh_t.REALS_t)
    for k, v in mesh_types.items():
        setattr(composite_mesh, k, np.array(
            [(cnn, 0) for cnn in np.vstack(elems[k])],
            dtype=getattr(jigsaw_msh_t, v)))

    composite_mesh.crs = dst_crs

    return composite_mesh



def add_pool_args(func):
    def wrapper(*args, nprocs=None, pool=None, **kwargs):

        # TODO: Modify docstring?
        if pool is not None:
            rv = func(*args, **kwargs, pool=pool)

        else:
            # Check nprocs
            nprocs = -1 if nprocs is None else nprocs
            nprocs = cpu_count() if nprocs == -1 else nprocs

            with Pool(processes=nprocs) as new_pool:
                rv = func(*args, **kwargs, pool=new_pool)
            new_pool.join()

        return rv

    return wrapper

def drop_extra_vertex_from_line(lstr: LineString) -> LineString:

    coords = np.array(lstr.coords)

    vecs = coords[1:] - coords[:-1]

    def is_zero(i):
        return np.isclose(i, 0)
    def isnt_zero(i):
        return np.logical_not(np.isclose(i, 0))

    vec_sizes = np.sqrt(vecs[:, 0] ** 2 + vecs[:, 1] ** 2)
    uvecs = vecs / vec_sizes.reshape(-1, 1)
    uvecs = uvecs[isnt_zero(vec_sizes)]
    nondup_coords = np.vstack([coords[:-1][isnt_zero(vec_sizes)], coords[-1]])

    vec_diffs = np.diff(uvecs, axis=0)
    diff_sizes = np.sqrt(vec_diffs[:, 0] ** 2 + vec_diffs[:, 1] ** 2)

    new_coords = np.vstack(
        [nondup_coords[0],
         nondup_coords[1:-1][isnt_zero(diff_sizes)],
         nondup_coords[-1]
         ])

    return LineString(new_coords)

def drop_extra_vertex_from_polygon(
        mpoly: Union[Polygon, MultiPolygon]) -> MultiPolygon:

    if isinstance(mpoly, Polygon):
        mpoly = MultiPolygon([mpoly])
    poly_seam_list = []
    for poly in mpoly.geoms:
        extr = drop_extra_vertex_from_line(poly.exterior)
        inters = [
            drop_extra_vertex_from_line(lstr)
            for lstr in poly.interiors]
        poly_seam_list.append(Polygon(extr, inters))

    return MultiPolygon(poly_seam_list)



from ocsmesh.ops.combine_geom import GeomCombine
from ocsmesh.ops.combine_hfun import HfunCombine

def combine_geometry(*args, **kwargs):
    return GeomCombine(*args, **kwargs).run()

def combine_hfun(*args, **kwargs):
    return HfunCombine(*args, **kwargs).run()

__all__ = [
        "combine_geometry",
        "combine_hfun"
]



import pathlib

from pyproj import CRS


def read(path, crs=None):
    sms2dm = {}
    with open(pathlib.Path(path), 'r') as f:
        lines = list(map(str.split, f.readlines()))
        ind = 1
        while ind < len(lines):
            line = lines[ind]
            ind = ind + 1
            if len(line) == 0:
                break
            if line[0] in ['E3T', 'E4Q']:
                if line[0] not in sms2dm:
                    sms2dm[line[0]] = {}
                v = []
                if line[0] == 'E3T':
                    v = line[2:5]
                elif line[0] == 'E4Q':
                    v = line[2:6]
                sms2dm[line[0]].update({
                    line[1]: v
                    })
            if line[0] == 'ND':
                if line[0] not in sms2dm:
                    sms2dm[line[0]] = {}
                sms2dm[line[0]].update({
                    line[1]: (
                        list(map(float, line[2:-1])), float(line[-1])
                        )
                    })
    if crs is not None:
        sms2dm['crs'] = CRS.from_user_input(crs)
    return sms2dm


def writer(sms2dm, path, overwrite=False):
    path = pathlib.Path(path)
    if path.is_file() and not overwrite:
        msg = 'File exists, pass overwrite=True to allow overwrite.'
        raise Exception(msg)
    with open(path, 'w') as f:
        f.write(to_string(sms2dm))


def to_string(sms2dm):
    data = ['MESH2D']
    E3T = E3T_string(sms2dm)
    if E3T is not None:
        data.append(E3T)
    E4Q = E4Q_string(sms2dm)
    if E4Q is not None:
        data.append(E4Q)
    data.append(ND_string(sms2dm))
    return '\n'.join(data)


def ND_string(sms2dm):
    assert all(int(nd_id) > 0 for nd_id in sms2dm['ND'])
    lines = []
    for nd_id, (coords, value) in sms2dm['ND'].items():
        lines.append(' '.join([
            'ND',
            f'{int(nd_id):d}',
            f"{coords[0]:<.16E}",
            f"{coords[1]:<.16E}",
            f"{value:<.16E}"
        ]))
    return '\n'.join(lines)


def geom_string(geom_type, sms2dm):
    assert geom_type in ['E3T', 'E4Q', 'E6T', 'E8Q', 'E9Q']
    assert all(int(elm_id) > 0 for elm_id in sms2dm[geom_type])
    f = []
    for elm_id, geom in sms2dm[geom_type].items():
        line = [
            f'{geom_type}',
            f'{elm_id}',
        ]
        for j, _ in enumerate(geom):
            line.append(f"{geom[j]}")
        f.append(' '.join(line))
    if len(f) > 0:
        return '\n'.join(f)

    return None


def E3T_string(sms2dm):
    return geom_string('E3T', sms2dm)


def E4Q_string(sms2dm):
    return geom_string('E4Q', sms2dm)



"""This module define class for mesh based size function
"""

import functools
import logging
import operator
from collections import defaultdict
from typing import Union, Optional
from multiprocessing import cpu_count, Pool
from time import time

from matplotlib.transforms import Bbox
from scipy.spatial import cKDTree
from jigsawpy import jigsaw_msh_t
import numpy as np
from pyproj import Transformer
from shapely import ops
from shapely.geometry import (
    LineString, MultiLineString, Polygon, MultiPolygon)

from ocsmesh.hfun.base import BaseHfun
from ocsmesh.crs import CRS as CRSDescriptor
from ocsmesh import Mesh
from ocsmesh import utils


_logger = logging.getLogger(__name__)

class HfunMesh(BaseHfun):
    """Mesh based size function.

    Creates a mesh based size function. The mesh size is specified
    at each point of the mesh based on the specified criteria.

    Attributes
    ----------
    hmin
    hmax
    mesh
    crs

    Methods
    -------
    msh_t()
        Return mesh sizes specified on the points of the  underlying
        mesh.
    size_from_mesh()
        Calculate values of the size function at each point on the
        underlying mesh, based on the length of edges connected to
        that point.
    add_patch(multipolygon, expansion_rate=None,
              target_size=None, nprocs=None)
        Add a region of fixed size refinement with optional expansion
        rate for points outside the region to achieve smooth size
        transition.
    add_feature(feature, expansion_rate, target_size=None,
                max_verts=200, *, nprocs=None, pool=None)
        Decorated method to add size refinement based on the specified
        `expansion_rate`, `target_size`, and distance from the input
        feature lines `feature`.
    get_bbox(**kwargs)
        Return  the bounding box of the underlying mesh.

    Notes
    -----
    Unlike raster size function, mesh based size function doesn't
    support constraint at this point.
    """

    _crs = CRSDescriptor()

    def __init__(self, mesh: Mesh) -> None:
        """Initialize a mesh based size function object

        Parameters
        ----------
        mesh : Mesh
            Input mesh object whose points are used for specifying
            sizes of the mesh to be generated. Note the underlying
            mesh is not copied, it is overriden by methods in the
            object of this type.

        Notes
        -----
        When a size function is created from a mesh, it takes
        the values associated with the underlying mesh. Unless
        `size_from_mesh` is called or refinements are applied, the
        values can be meaningless.

        Unlike raster size function where the user defines the
        minimum and maximum, for mesh based hfun, the minimum and
        maximum is based on the values stored on the mesh.

        Note that currently object of this type holds onto a crs
        variable separate from `mesh.crs`. Because of this, if the
        user is not careful, one might run into unexpected behavior!
        """

        self._mesh = mesh
        self._crs = mesh.crs

    def msh_t(self) -> jigsaw_msh_t:
        """Return the size function specified on the underlying mesh

        Return the size function values stored on the underlying mesh.
        The return value is in a projected CRS. If the input mesh
        CRS is geographic, then a local UTM CRS is calculated and used
        for the output of this method.

        Parameters
        ----------

        Returns
        -------
        jigsaw_msh_t
            The size function specified on the points of input mesh.

        Notes
        -----
        This method effectively overrides the CRS of the objects and
        modifies it if the CRS is initially geographic. Note that
        this also affects the underlying mesh object which is **not**
        copied in the contructor.
        """

        utm_crs = utils.estimate_mesh_utm(self.mesh.msh_t)
        if utm_crs is not None:
            transformer = Transformer.from_crs(
                self.crs, utm_crs, always_xy=True)
            # TODO: This modifies the underlying mesh, is this
            # intended?
            self.mesh.msh_t.vert2['coord'] = np.vstack(
                transformer.transform(
                    self.mesh.msh_t.vert2['coord'][:, 0],
                    self.mesh.msh_t.vert2['coord'][:, 1]
                    )).T
            self.mesh.msh_t.crs = utm_crs
            self._crs = utm_crs

        return self.mesh.msh_t

    def size_from_mesh(self) -> None:
        """Calculates sizes based on the underlying mesh edge lengths.

        Get size function values based on the underlying input mesh
        This method overwrites the values in underlying `msh_t`.

        Parameters
        ----------

        Returns
        -------
        None

        Notes
        -----
        The size calculations are done in a projected CRS or a local
        UTM CRS. However, this method does not modify the
        CRS of the size function, it only updates the values and
        discard the projected CRS it used for calculations.


        """

        # Make sure it's in utm so that sizes are in meters
        hfun_msh = self.mesh.msh_t
        coord = hfun_msh.vert2['coord']

        transformer = None
        utm_crs = utils.estimate_mesh_utm(hfun_msh)
        if utm_crs is not None:
            _logger.info('Projecting to utm...')

            transformer = Transformer.from_crs(
                self.crs, utm_crs, always_xy=True)

        # Calculate length of all edges based on acquired coords
        _logger.info('Getting length of edges...')
        len_dict = utils.calculate_edge_lengths(hfun_msh, transformer)

        # Calculate the mesh size by getting average of lengths
        # associated with each vertex (note there's not id vs index
        # distinction here). This is the most time consuming section
        # as of 04/21
        vert_to_lens = defaultdict(list)
        for verts_idx, edge_len in len_dict.items():
            for vidx in verts_idx:
                vert_to_lens[vidx].append(edge_len)

        _logger.info('Creating size value array for vertices...')
        vert_value = np.array(
            [np.average(vert_to_lens[i]) if i in vert_to_lens else 0
             for i in range(coord.shape[0])])

        # NOTE: Modifying values of underlying mesh
        hfun_msh.value = vert_value.reshape(len(vert_value), 1)


    def add_patch(
            self,
            multipolygon: Union[MultiPolygon, Polygon],
            expansion_rate: Optional[float] = None,
            target_size: Optional[float] = None,
            nprocs: Optional[int] = None
            ) -> None:
        """Add refinement as a region of fixed size with an optional rate

        Add a refinement based on a region specified by `multipolygon`.
        The fixed `target_size` refinement can be expanded outside the
        region specified by the shape if `expansion_rate` is provided.

        Parameters
        ----------
        multipolygon : MultiPolygon or Polygon
            Shape of the region to use specified `target_size` for
            refinement.
        expansion_rate : float or None, default=None
            Optional rate to use for expanding refinement outside
            the specified shape in `multipolygon`.
        target_size : float or None, default=None
            Fixed target size of mesh to use for refinement in
            `multipolygon`
        nprocs : int or None, default=None
            Number of processors to use in parallel sections of the
            algorithm

        Returns
        -------
        None

        See Also
        --------
        add_feature :
            Add refinement for specified line string
        """

        # TODO: Add pool input support like add_feature for performance

        # TODO: Support other shapes - call buffer(1) on non polygons(?)
        if not isinstance(multipolygon, (Polygon, MultiPolygon)):
            raise TypeError(
                    f"Wrong type \"{type(multipolygon)}\""
                    f" for multipolygon input.")

        if isinstance(multipolygon, Polygon):
            multipolygon = MultiPolygon([multipolygon])

        # Check nprocs
        nprocs = -1 if nprocs is None else nprocs
        nprocs = cpu_count() if nprocs == -1 else nprocs
        _logger.debug(f'Using nprocs={nprocs}')


        # check target size
        target_size = self.hmin if target_size is None else target_size
        if target_size is None:
            # TODO: Is this relevant for mesh type?
            raise ValueError('Argument target_size must be specified if no '
                             'global hmin has been set.')
        if target_size <= 0:
            raise ValueError("Argument target_size must be greater than zero.")

        # For expansion_rate
        if expansion_rate is not None:
            exteriors = [ply.exterior for ply in multipolygon]
            interiors = [
                inter for ply in multipolygon for inter in ply.interiors]

            features = MultiLineString([*exteriors, *interiors])
            # pylint: disable=E1123, E1125
            self.add_feature(
                feature=features,
                expansion_rate=expansion_rate,
                target_size=target_size,
                nprocs=nprocs)

        coords = self.mesh.msh_t.vert2['coord']
        values = self.mesh.msh_t.value

        verts_in = utils.get_verts_in_shape(
            self.mesh.msh_t, shape=multipolygon, from_box=False)

        if len(verts_in):
            # NOTE: Don't continue, otherwise the final
            # destination file might end up being empty!
            values[verts_in, :] = target_size

        # NOTE: unlike raster self.hmin is based on values of this
        # hfun before applying feature; it is ignored so that
        # the new self.hmin becomes equal to "target" specified
#        if self.hmin is not None:
#            values[np.where(values < self.hmin)] = self.hmin
        if self.hmax is not None:
            values[np.where(values > self.hmax)] = self.hmax
        values = np.minimum(self.mesh.msh_t.value, values)
        values = values.reshape(self.mesh.msh_t.value.shape)

        self.mesh.msh_t.value = values

    @utils.add_pool_args
    def add_feature(
            self,
            feature: Union[LineString, MultiLineString],
            expansion_rate: float,
            target_size: float = None,
            max_verts=200,
            *, # kwarg-only comes after this
            pool: Pool
    ):
        """Add refinement for specified linestring `feature`

        Add refinement for the specified linestrings `feature`.
        The refinement is relaxed with `expansion_rate` and distance
        from the extracted contour lines.

        Parameters
        ----------
        feature : LineString or MultiLineString
            The user-specified line strings for applying refinement on.
        expansion_rate : float
            Rate to use for expanding refinement with distance away
            from the extracted contours.
        target_size : float or None, default=None
            Target size to use on the extracted contours and expand
            from with distance.
        max_verts : int, default=200
            Number of maximum vertices in a feature line that is
            passed to a separate process in parallel section of
            the algorithm.
        pool : Pool
            Pre-created and initialized process pool to be used for
            parallel sections of the algorithm.

        Returns
        -------
        None

        See Also
        --------
        add_patch :
            Add refinement for region specified polygon

        Notes
        -----
        See https://outline.com/YU7nSM for an explanation
        about tree algorithms.

        Creating a local projection allows having similar area/length
        calculations as if great circle calculations was being used.

        Another useful refererence:
        https://gis.stackexchange.com/questions/214261/should-we-always-calculate-length-and-area-in-lat-lng-to-get-accurate-sizes-leng
        """
        # TODO: Partition features if they are too "long" which results in an
        # improvement for parallel pool. E.g. if a feature is too long, 1
        # processor will be busy and the rest will be idle.

        if not isinstance(feature, (LineString, MultiLineString)):
            raise TypeError(
                f'Argument feature must be of type {LineString} or '
                f'{MultiLineString}, not type {type(feature)}.')

        if isinstance(feature, LineString):
            feature = [feature]

        elif isinstance(feature, MultiLineString):
            feature = list(feature)

        # check target size
        target_size = self.hmin if target_size is None else target_size
        if target_size is None:
            raise ValueError('Argument target_size must be specified if no '
                             'global hmin has been set.')
        if target_size <= 0:
            raise ValueError("Argument target_size must be greater than zero.")

        utm_crs = utils.estimate_mesh_utm(self.mesh.msh_t)

        _logger.info('Repartitioning features...')
        start = time()
        res = pool.starmap(
            utils.repartition_features,
            [(linestring, max_verts) for linestring in feature]
            )
        feature = functools.reduce(operator.iconcat, res, [])
        _logger.info(f'Repartitioning features took {time()-start}.')

        _logger.info('Resampling features on ...')
        start = time()

        # We don't want to recreate the same transformation
        # many times (it takes time) and we can't pass
        # transformation object to subtask (cinit issue)
        transformer = None
        if utm_crs is not None:
            start2 = time()
            transformer = Transformer.from_crs(
                self.crs, utm_crs, always_xy=True)
            _logger.info(
                    f"Transform creation took {time() - start2:f}")
            start2 = time()
            feature = [
                ops.transform(transformer.transform, linestring)
                for linestring in feature]
            _logger.info(
                    f"Transform apply took {time() - start2:f}")

        transformed_features = pool.starmap(
            utils.transform_linestring,
            [(linestring, target_size) for linestring in feature]
        )
        _logger.info(f'Resampling features took {time()-start}.')
        _logger.info('Concatenating points...')
        start = time()
        points = []
        for geom in transformed_features:
            if isinstance(geom, LineString):
                points.extend(geom.coords)
            elif isinstance(geom, MultiLineString):
                for linestring in geom:
                    points.extend(linestring.coords)
        _logger.info(f'Point concatenation took {time()-start}.')

        _logger.info('Generating KDTree...')
        start = time()
        tree = cKDTree(np.array(points))
        _logger.info(f'Generating KDTree took {time()-start}.')

        # We call msh_t() so that it also takes care of utm
        # transformation
        xy = self.msh_t().vert2['coord']

        _logger.info(f'transforming points took {time()-start}.')
        _logger.info('querying kdtree...')
        start = time()
        if self.hmax:
            r = (self.hmax - target_size) / (expansion_rate * target_size)
            near_dists, neighbors = tree.query(
                xy, workers=pool._processes, distance_upper_bound=r)
            distances = r * np.ones(len(xy))
            mask = np.logical_not(np.isinf(near_dists))
            distances[mask] = near_dists[mask]
        else:
            distances, _ = tree.query(xy, workers=pool._processes)
        _logger.info(f'querying kdtree took {time()-start}.')
        values = expansion_rate*target_size*distances + target_size
        # NOTE: unlike raster self.hmin is based on values of this
        # hfun before applying feature; it is ignored so that
        # the new self.hmin becomes equal to "target" specified
#        if self.hmin is not None:
#            values[np.where(values < self.hmin)] = self.hmin
        if self.hmax is not None:
            values[np.where(values > self.hmax)] = self.hmax
        values = np.minimum(self.mesh.msh_t.value.ravel(), values)
        values = values.reshape(self.mesh.msh_t.value.shape)

        self.mesh.msh_t.value = values

    @property
    def hmin(self):
        """Read-only attribute for the minimum mesh size constraint"""

        return np.min(self.mesh.msh_t.value)

    @property
    def hmax(self):
        """Read-only attribute for the maximum mesh size constraint"""

        return np.max(self.mesh.msh_t.value)

    @property
    def mesh(self):
        """Read-only attribute to reference to the input mesh"""

        return self._mesh

    @property
    def crs(self):
        """Read-only attribute holding onto hfun CRS"""

        return self._crs

    def get_bbox(self, **kwargs) -> Union[Polygon, Bbox]:
        """Returns the bounding box of the underlying mesh

        Parameters
        ----------
        kwargs : dict, optional
            Arguments passed to the underlying mesh `get_bbox`

        Returns
        -------
        Polygon or Bbox
            The bounding box of the underlying mesh
        """

        return self.mesh.get_bbox(**kwargs)



"""Wrapper for raster image functionalities commonly used for meshing

This module implements wrapper for basic functionalities of handling
raster files such as CRS transformations, resampling, clipping,
extracting & overriding data, plotting, etc.
"""

import math
import hashlib
import logging
import multiprocessing
import os
import pathlib
import tempfile
import warnings
from time import time
from contextlib import contextmanager, ExitStack
from typing import (
        Union, Generator, Any, Optional, Dict, List, Tuple, Iterable,
        Callable)
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import geopandas as gpd
import matplotlib.pyplot as plt
from matplotlib.cm import ScalarMappable
from matplotlib.transforms import Bbox
from matplotlib.axes import Axes
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
import numpy.typing as npt
from numpy import ma
from pyproj import CRS, Transformer
import rasterio
import rasterio.mask
from rasterio import warp, Affine
from rasterio.enums import Resampling
from rasterio.fill import fillnodata
from rasterio.transform import array_bounds
from rasterio import windows
from scipy.ndimage import gaussian_filter, generic_filter
from scipy import LowLevelCallable
from shapely import ops
from shapely.geometry import (
    Polygon, MultiPolygon, LineString, MultiLineString, box)
from numba import cfunc, carray
from numba.types import intc, intp, float64, voidptr
from numba.types import CPointer

from ocsmesh import figures
from ocsmesh import utils

_logger = logging.getLogger(__name__)


tmpdir = str(pathlib.Path(tempfile.gettempdir()+'/ocsmesh'))+'/'
os.makedirs(tmpdir, exist_ok=True)


# From https://ilovesymposia.com/2017/03/12/scipys-new-lowlevelcallable-is-a-game-changer/
@cfunc(intc(CPointer(float64), intp,
            CPointer(float64), voidptr))
def nbmean(values_ptr, len_values, result, data):
    values = carray(values_ptr, (len_values,), dtype=float64)
    result[0] = 0
    n_vals = 0
    for v in values:
        if not np.isnan(v):
            result[0] = result[0] + v
            n_vals = n_vals + 1
    # NOTE: Divide by number of used values, NOT the whole foot area
    if n_vals > 1:
        result[0] = result[0] / n_vals

    return 1


class RasterPath:
    """Descriptor class for storing the path of original input raster
    """

    def __set__(self, obj, val: Union[str, os.PathLike]):
        obj.__dict__['path'] = pathlib.Path(val)

    def __get__(self, obj, objtype=None):
        return obj.__dict__['path']


class Crs:
    """Descriptor class for CRS of the input raster.

    If user specifies this value, the existing CRS of the input raster
    is ignored (if any). If the user doesn't provide this value and
    the raster itself doesn't have any CRS either it's an error.
    """

    def __set__(self, obj, val: Union[str, CRS, None]):

        # check if CRS is in file
        if val is None:
            with rasterio.open(obj.path) as src:
                # Raise if CRS not in file and the user did not provide a CRS.
                # All Rasters objects must have a defined CRS.
                # Program cannot operate with an undefined CRS.
                val = src.crs
                if val is None:
                    raise IOError(
                        'CRS not found in raster file. Must specify CRS.')
        # CRS is specified by user rewrite raster but add CRS to meta
        else:
            if isinstance(val, str):
                val = CRS.from_user_input(val)

            if not isinstance(val, CRS):
                raise TypeError(f'Argument crs must be of type {str} or {CRS},'
                                f' not type {type(val)}.')

            # create a temporary copy of the original file and update meta.
            with ExitStack() as stack:

                src = stack.enter_context(rasterio.open(obj.path))
                if obj.chunk_size is not None:
                    wins = get_iter_windows(
                        src.width, src.height, chunk_size=obj.chunk_size)
                else:
                    wins = [windows.Window(
                        0, 0, src.width, src.height)]

                dst = stack.enter_context(
                        obj.modifying_raster(crs=val, driver='GTiff'))
                for window in wins:
                    dst.write(src.read(window=window), window=window)


class TemporaryFile:
    """Descriptor class for storing tempfile object of working raster

    All the raster operations done using `Raster` object is written
    to disk. In order to avoid modifying the original input raster
    a the input data is first writton to a temporary file. This
    temporary file is created using `TemporaryFile` to have auto
    cleanup capabities on object destruction.
    """

    def __set__(self, obj, val: tempfile.NamedTemporaryFile):
        obj.__dict__['tmpfile'] = val
        obj._src = rasterio.open(val.name)

    def __get__(self, obj, objtype=None) -> pathlib.Path:
        tmpfile = obj.__dict__.get('tmpfile')
        if tmpfile is None:
            return obj.path
        return pathlib.Path(tmpfile.name)


class SourceRaster:
    """Descriptor class for storing the raster dataset handler.

    `Raster` objects are manipulated using `rasterio` package. All
    the access to the raster data using this package happens through
    `rasterio.DatasetReader`. This descriptor holds onto the opened
    dataset to use as a shorthand of checking file existance and
    opening it everytime need arises.
    """

    def __set__(self, obj, val: rasterio.DatasetReader):
        obj.__dict__['source'] = val

    def __get__(self, obj, objtype=None) -> rasterio.DatasetReader:
        source = obj.__dict__.get('source')
        if source is None:
            source = rasterio.open(obj.path)
            obj.__dict__['source'] = source
        return source


class ChunkSize:
    """Descriptor class for storing the size for windowed operations
    """

    def __set__(self, obj, val: int):
        chunk_size = 0 if val is None else int(val)
        if not chunk_size >= 0:
            raise ValueError("Argument chunk_size must be >= 0.")
        obj.__dict__['chunk_size'] = val

    def __get__(self, obj, objtype=None) -> int:
        return obj.__dict__['chunk_size']


class Overlap:
    """Descriptor class for storing the overlap for windowed operations
    """

    def __set__(self, obj, val: int):
        obj.__dict__['overlap'] = 0 if val is None else val

    def __get__(self, obj, objtype=None) -> int:
        return obj.__dict__['overlap']


class Raster:
    """Wrapper class for basic raster handling

    Attributes
    ----------
    x
    y
    values
    path
    tmpfile
    md5
    count
    is_masked
    shape
    height
    bbox
    src
    width
    dx
    dy
    crs
    nodatavals
    transform
    dtypes
    nodata
    xres
    yres
    resampling_method
    chunk_size
    overlap

    Methods
    -------
    modifying_raster(use_src_meta=True, **kwargs)
        Context for modifying raster and saving to disk after.
    get_x(window=None)
        Get X values from the raster.
    get_y(window=None)
        Get Y values from the raster.
    get_xy(window=None)
        Get raster position tuples.
    get_values(window=None, band=None, **kwargs)
        Get values at all the points in the raster.
    get_xyz(window=None, band=None)
        Get raster position tuples and values horizontally stacked.
    get_multipolygon(zmin=None, zmax=None, window=None, overlap=None, band=1)
        Extract multipolygon from raster data.
    get_bbox(crs=None, output_type='polygon')
        Get the raster bounding box.
    contourf(...)
        Plot a filled contour from the raster data.
    tags(i=None)
        Get tags set on raster dataset.
    read(i, masked=True, **kwargs)
        Call underlying dataset `read` method.
    dtype(i)
        Return data type from the source raster.
    nodataval(i)
        Return the no-data value of the source raster.
    sample(xy, i)
        Call underlying dataset `sample` method.
    close()
        Release source raster object.
    add_band(values,  **tags)
        Add a new band of data with specified tags to the raster.
    fill_nodata()
        Fill no-data points in the raster dataset.
    gaussian_filter(**kwargs)
        Apply Gaussian filter on the raster data.
    average_filter(size, drop_above, drop_below)
        Apply average filter on the raster data.
    generic_filter(function, **kwargs)
        Apply generic filter on the raster data.
    mask(shapes, i=None, **kwargs)
        Mask raster data by shape.
    read_masks(i=None)
        Read source raster masks.
    warp(dst_crs, nprocs=-1)
        Reproject raster data.
    resample(scaling_factor, resampling_method=None)
        Resample raster data.
    save(path)
        Save-as raster data to the provided path.
    clip(geom)
        Clip raster data by provided shape.
    adjust(geom=None, inside_min=-np.inf, outside_max=np.inf, cond=None)
        Modify raster values based on constraints and shape.
    get_contour(level, window)
        Calculate contour of specified level on raster data.
    get_channels(level=0, width=1000, tolerance=None)
        Calculate narrow areas based on input level and width.
    iter_windows(chunk_size=None, overlap=None)
        Return raster view windows based on chunk size and overlap.
    get_window_data(window, masked=True, band=None)
        Return raster values based for the input window.
    get_window_bounds(window)
        Return window bounds.
    get_window_transform(window)
        Return window transform for the input raster

    Notes
    -----
    This class makes use of property and descriptor type attributes.
    Some of the properties even point to the value set by the
    descriptors. It's important to not that the properties are
    set lazily. It's also noteworthy to mention that this class is
    currently **not** picklable due temporary file and file-like
    object attributes.
    """

    _path = RasterPath()
    _crs = Crs()
    _chunk_size = ChunkSize()
    _overlap = Overlap()
    _tmpfile = TemporaryFile()
    _src = SourceRaster()

    def __init__(
            self,
            path: Union[str, os.PathLike],
            crs: Union[str, CRS, None] = None,
            chunk_size: Optional[int] = None,
            overlap: Optional[int] = None
    ):
        """Raster data manipulator.

        Parameters
        ----------
        path : str or os.PathLike
            Path to a raster image to work on (.tiff or .nc).
        crs : str or CRS, default=None
            CRS to use and override input raster data with.
            Note that no transformation takes place.
        chunk_size : int or None, default=None
            Square window size to be used for data chunking
        overlap : int or None , default=None
            Overlap size for calculating chunking windows on the raster
        """

        self._chunk_size = chunk_size
        self._overlap = overlap
        self._path = path
        self._crs = crs

    def __iter__(self, chunk_size: int = None, overlap: int = None):
        for window in self.iter_windows(chunk_size, overlap):
            yield window, self.get_window_bounds(window)

    @contextmanager
    def modifying_raster(
            self,
            use_src_meta: bool = True,
            **kwargs: Any
            ) -> Generator[rasterio.DatasetReader, None, None]:
        r"""Context manager for modifying and storing raster data

        This is a helper context manager method that handles creating
        new temporary file and replacing the old one with it when
        raster data is successfully modified.

        Parameters
        ----------
        use_src_meta : bool, default=True
            Whether or not to copy the metadata of the source raster
            when creating the new empty raster file
        **kwargs : dict, optional
            Options to be passed as metadata to raster database. These
            options override values taken from source raster in case
            `use_src_meta` is `True`

        Yields
        ------
        rasterio.DatasetReader
            Handle to the opened dataset on temporary file which
            will override the old values if no exception occurs
            during the context
        """

        no_except = False
        try:
            # pylint: disable=R1732
            tmpfile = tempfile.NamedTemporaryFile(prefix=tmpdir)

            new_meta = kwargs
            # Flag to workaround cases where "src" is NOT set yet
            if use_src_meta:
                new_meta = self.src.meta.copy()
                new_meta.update(**kwargs)
            with rasterio.open(tmpfile.name, 'w', **new_meta) as dst:
                if use_src_meta:
                    for i, desc in enumerate(self.src.descriptions):
                        dst.set_band_description(i+1, desc)
                yield dst

            no_except = True

        finally:
            if no_except:
                # So that tmpfile is NOT destroyed when it locally
                # goes out of scope
                self._tmpfile = tmpfile



    def get_x(
            self,
            window: Optional[windows.Window] = None
            ) -> npt.NDArray[float]:
        """Get X positions of the raster grid.

        Parameters
        ----------
        window : windows.Window, default=None
            The window over which X positions are to be returned

        Returns
        -------
        np.ndarray
            A vector X positions from minimum to the maximum
        """

        window = windows.Window(0, 0, self.src.shape[1], self.src.shape[0]) \
            if window is None else window
        if window is not None:
            assert isinstance(window, windows.Window)
            width = window.width
        else:
            width = self.shape[0]
        x0, y0, x1, y1 = self.get_window_bounds(window)
        return np.linspace(x0, x1, width)

    def get_y(
            self,
            window: Optional[windows.Window] = None
            ) -> npt.NDArray[float]:
        """Get Y positions of the raster grid.

        Parameters
        ----------
        window : windows.Window, default=None
            The window over which Y positions are to be returned

        Returns
        -------
        np.ndarray
            A vector Y positions from minimum to the maximum
        """

        window = windows.Window(0, 0, self.src.shape[1], self.src.shape[0]) \
            if window is None else window
        if window is not None:
            assert isinstance(window, windows.Window)
            height = window.height
        else:
            height = self.shape[1]
        x0, y0, x1, y1 = self.get_window_bounds(window)
        return np.linspace(y1, y0, height)

    def get_xy(
            self,
            window: Optional[windows.Window] = None
            ) -> np.ndarray:
        """Get raster positions tuple array

        Parameters
        ----------
        window : windows.Window, default=None
            The window over which positions are to be returned

        Returns
        -------
        np.ndarray
            A `n` :math:`\times 2` matrix of positions
        """

        x, y = np.meshgrid(self.get_x(window), self.get_y(window))
        return np.vstack([x.flatten(), y.flatten()]).T

    def get_values(
            self,
            window: Optional[windows.Window] = None,
            band: Optional[int] = None,
            **kwargs: Any
            ) -> npt.NDArray[float]:
        r"""Return the data stored at each point in the raster grid

        Parameters
        ----------
        window : windows.Window, default=None
            The window over which values are to be returned.
        band : int, default=None
            The band from which the values should be read. If `None`
            return data from band `1`.
        **kwargs : dict, optional
            Additional arguments to pass to `rasterio.DatasetReader.read`.

        Returns
        -------
        np.ndarray
            A 2D matrix of values on the raster grid.
        """

        i = 1 if band is None else band
        window = windows.Window(0, 0, self.src.shape[1], self.src.shape[0]) \
            if window is None else window
        if window is not None:
            assert isinstance(window, windows.Window)
        return self.src.read(i, window=window, **kwargs)

    def get_xyz(
            self,
            window: Optional[windows.Window] = None,
            band: Optional[int] = None
            ) -> np.ndarray:
        """Return the data stored at each point in the raster grid

        Parameters
        ----------
        window : windows.Window, default=None
            The window over which positions and values are to be
            returned.
        band : int, default=None
            The band from which the values should be read. If `None`
            return data from band `1`.

        Returns
        -------
        np.ndarray
            A `n` :math:`\times 3` matrix of x, y and values
        """

        xy = self.get_xy(window)
        values = self.get_values(window=window, band=band).reshape(
            (xy.shape[0], 1))
        return np.hstack([xy, values])

    def get_multipolygon(
            self,
            zmin: Optional[float] = None,
            zmax: Optional[float] = None,
            window: Optional[windows.Window] = None,
            overlap: Optional[int] = None,
            band: int = 1,
    ) -> MultiPolygon:
        """Calculate and return a multipolygon based on the raster data

        Calculates filled contour from raster data between specified
        limits on the specified band and creates a multipolygon from
        the filled contour to return.

        Parameters
        ----------
        zmin : float or None, default=None
            Lower bound of raster data for filled contour calculation
        zmax : float or None, default=None
            Upper bound of raster data for filled contour calculation
        window : windows.Window or None, default=None
            Window over whose data the multipolygon is calculated
        overlap : int or None, default=None
            Overlap used for generating windows if `window` is not provided
        band : int, default=1
            Raster band over whose data multipolygon is calculated

        Returns
        -------
        MultiPolygon
            The calculated multipolygon from raster data
        """

        polygon_collection = []
        if window is None:
            iter_windows = self.iter_windows(overlap=overlap)
        else:
            iter_windows = [window]

        for win in iter_windows:
            x, y, z = self.get_window_data(win, band=band)
            new_mask = np.full(z.mask.shape, 0)
            new_mask[np.where(z.mask)] = -1
            new_mask[np.where(~z.mask)] = 1

            if zmin is not None:
                new_mask[np.where(z < zmin)] = -1

            if zmax is not None:
                new_mask[np.where(z > zmax)] = -1

            if np.all(new_mask == -1):  # or not new_mask.any():
                continue

            fig, ax = plt.subplots()
            ax.contourf(x, y, new_mask, levels=[0, 1])
            plt.close(fig)
            polygon_collection.extend(
                    utils.get_multipolygon_from_pathplot(ax).geoms)

        union_result = ops.unary_union(polygon_collection)
        if not isinstance(union_result, MultiPolygon):
            union_result = MultiPolygon([union_result])
        return union_result

    def get_bbox(
            self,
            crs: Union[str, CRS, None] = None,
            output_type: Literal['polygon', 'bbox'] = 'polygon'
        ) -> Union[Polygon, Bbox]:
        """Calculate the bounding box of the raster.

        Parameters
        ----------
        crs : str or CRS or None, default=None
            The CRS in which the bounding box is requested.
        output_type : {'polygon', 'bbox'}
            The label of the return type for bounding box, either a
            `shapely` 'polygon' or `matplotlib` 'bbox'.

        Returns
        -------
        box or Bbox
            The bounding box of the raster.

        Raises
        ------
        TypeError
            If the label of return type is not valid.
        """

        output_type = 'polygon' if output_type is None else output_type
        xmin, xmax = np.min(self.x), np.max(self.x)
        ymin, ymax = np.min(self.y), np.max(self.y)
        crs = self.crs if crs is None else crs
        if crs is not None:
            if not self.crs.equals(crs):
                transformer = Transformer.from_crs(
                    self.crs, crs, always_xy=True)
                # pylint: disable=E0633
                (xmin, xmax), (ymin, ymax) = transformer.transform(
                    (xmin, xmax), (ymin, ymax))
        if output_type == 'polygon': # pylint: disable=R1705
            return box(xmin, ymin, xmax, ymax)
        elif output_type == 'bbox':
            return Bbox([[xmin, ymin], [xmax, ymax]])

        raise TypeError(
            'Argument output_type must a string literal \'polygon\' or '
            '\'bbox\'')

    def contourf(
            self,
            band: int = 1,
            window: Optional[windows.Window] = None,
            axes: Optional[Axes] = None,
            vmin: Optional[float] = None,
            vmax: Optional[float] = None,
            cmap: str = 'topobathy',
            levels: Optional[List[float]] = None,
            show: bool = False,
            title: Optional[str] = None,
            figsize: Optional[Tuple[float, float]] = None,
            colors: int = 256,
            cbar_label: Optional[str] = None,
            norm=None,
            **kwargs : Any
            ) -> Axes:
        """Plot filled contour for raster data.

        Parameters
        ----------
            band : int, default=1
                Raster band from which data is used.
            window : windows.Window or None, default=None
                Raster window from which data is used.
            axes : Axes or None, default=None
                Matplotlib axes to draw contour on>
            vmin : float or None, default=None
                Minimum value of the filled contour.
            vmax : float or None, default=None
                Maximum value of the filled contour.
            cmap : str, default='topobathy'
                Colormap to use for filled contour.
            levels : list of float or None, default=None
                Prespecified list of contour levels.
            show : bool, default=False
                Whether to show the contour on creation or not.
            title : str or None, default=None
                Title used on the axes of the contour
            figsize : tuple of float or None, default=None
                Figure size used for the contour figure
            colors : int, default=256
                Contour colors associated with levels
            cbar_label : str or None, default=None
                Label of the colorbar
            norm : Normalize or None, default=None
                Normalizer object
            **kwargs : dict, optional
                Keyword arguments passed to the matplotlib contourf()
                function

        Returns
        -------
        Axes
            Axes object from matplotlib library that holds onto the
            contour plot object
        """

        if axes is None:
            fig = plt.figure(figsize=figsize)
            axes = fig.add_subplot(111)
        values = self.get_values(band=band, masked=True, window=window)
        vmin = np.min(values) if vmin is None else float(vmin)
        vmax = np.max(values) if vmax is None else float(vmax)
        cmap, norm, levels, col_val = figures.get_topobathy_kwargs(
            values, vmin, vmax)
        axes.contourf(
            self.get_x(window),
            self.get_y(window),
            values,
            levels=levels,
            cmap=cmap,
            norm=norm,
            vmin=vmin,
            vmax=vmax,
            **kwargs
            )
        axes.axis('scaled')
        if title is not None:
            axes.set_title(title)
        mappable = ScalarMappable(cmap=cmap)
        mappable.set_array([])
        mappable.set_clim(vmin, vmax)
        divider = make_axes_locatable(axes)
        cax = divider.append_axes("bottom", size="2%", pad=0.5)
        cbar = plt.colorbar(
            mappable,
            cax=cax,
            # extend=cmap_extend,
            orientation='horizontal'
            )
        if col_val != 0:
            cbar.set_ticks([vmin, vmin + col_val * (vmax-vmin), vmax])
            cbar.set_ticklabels([np.around(vmin, 2), 0.0, np.around(vmax, 2)])
        else:
            cbar.set_ticks([vmin, vmax])
            cbar.set_ticklabels([np.around(vmin, 2), np.around(vmax, 2)])
        if cbar_label is not None:
            cbar.set_label(cbar_label)
        if show is True:
            plt.show()
        return axes

    def tags(self, i: Optional[int] = None) -> Dict[str, str]:
        """Return a dictionary of dataset or band's tags

        Parameters
        ----------
        i : int or None, default=None
            The band from which the tags are read.

        Returns
        -------
        dict
            Dictionary of tags
        """

        if i is None:
            return self.src.tags()
        return self.src.tags(i)

    def read(self, i: int, masked: bool = True, **kwargs: Any) -> npt.NDArray[float]:
        """Read the data from raster opened file

        Parameters
        ----------
        i : int
            The index of the band to read the data from
        masked : bool, default=True
            Whether or not to return a masked array
        **kwargs : dict, optional
            Additional keyword arguments passed to rasterio read()

        Returns
        -------
        ndarray
            Array of raster data
        """

        return self.src.read(i, masked=masked, **kwargs)

    def dtype(self, i: int) -> npt.NDArray[float]:
        """Raster data type

        Parameters
        ----------
        i : int
            The index of the band to read the data from

        Returns
        -------
        Any
            Data type of raster values
        """

        return self.src.dtypes[i-1]

    def nodataval(self, i: int) -> float:
        """Value used for filling no-data points

        Parameters
        ----------
        i : int
            The index of the band to read the data from

        Returns
        -------
        float
            The value to be used for points that have missing data
        """

        return self.src.nodatavals[i-1]

    def sample(self, xy: Iterable, i: int) -> npt.NDArray[float]:
        """Get value of the data in specified positions

        Parameters
        ----------
        xy : iterable
            Pairs of xy coordinates for which data is retrieved
        i : int
            The index of the band to read the data from

        Returns
        -------
        ndarray
            Array of values for specified input positions
        """

        return self.src.sample(xy, i)

    def close(self) -> None:
        """Delete source object"""

        del self._src

    def add_band(self, values: npt.NDArray[float], **tags: Any) -> int:
        """Add a new band for `values` with tags `tags` to the raster

        Parameters
        ----------
            values : array-like
                The values to be added to the raster, it must have the
                correct shape as the raster.
            **tags : dict, optional
                The tags to be added for the new band of data.

        Returns
        -------
        int
            ID of the new band added to the raster.
        """

        kwargs = self.src.meta.copy()
        band_id = kwargs["count"] + 1
        with self.modifying_raster(count=band_id) as dst:
            for i in range(1, self.src.count + 1):
                dst.write_band(i, self.src.read(i))
            dst.write_band(band_id, values.astype(self.src.dtypes[i-1]))
        return band_id

    def fill_nodata(self) -> None:
        """Fill missing values in the raster in-place"""

        # A parallelized version is presented here:
        # https://github.com/basaks/rasterio/blob/master/examples/fill_large_raster.py

        with self.modifying_raster() as dst:
            for window in self.iter_windows():
                dst.write(
                    fillnodata(self.src.read(window=window, masked=True)),
                    window=window
                    )

    def gaussian_filter(self, **kwargs: Any) -> None:
        """Apply Gaussian filter to the raster data in-place

        Parameters
        ----------
        **kwargs : dict, optional
            Keyword arguments passed to SciPy `gaussian_filter` function

        Returns
        -------
        None
        """

        # TODO: Don't overwrite; add additoinal bands for filtered values

        # NOTE: Adding new bands in this function can result in issues
        # in other parts of the code. Thorough testing is needed for
        # modifying the raster (e.g. hfun add_contour is affected)
        with self.modifying_raster() as dst:
            for i in range(1, self.src.count + 1):
                outband = self.src.read(i, masked=True)
#                # Write orignal band
#                dst.write_band(i + n_bands_new // 2, outband)
                # Write filtered band
                outband = gaussian_filter(outband, **kwargs)
                dst.write_band(i, outband)

    def average_filter(
            self,
            size: Union[int, npt.NDArray[int]],
            drop_above: Optional[float] = None,
            drop_below: Optional[float] = None,
            apply_on_bands: Optional[List[int]] = None
            ) -> None:
        """Apply average(mean) filter on the raster

        Parameters
        ----------
        size: int, npt.NDArray[int]
            size of the footprint
        drop_above: float or None
            elevation above which the cells are ignored for averaging
        drop_below: float or None
            elevation below which the cells are ignored for averaging

        Returns
        -------
        None
        """

        # TODO: Don't overwrite; add additoinal bands for filtered values

        # NOTE: Adding new bands in this function can result in issues
        # in other parts of the code. Thorough testing is needed for
        # modifying the raster (e.g. hfun add_contour is affected)

        bands = apply_on_bands
        if bands is None:
            bands = range(1, self.src.count + 1)
        with self.modifying_raster() as dst:
            for i in bands:
                bnd_idx = i - 1
                outband = self.src.read(i, masked=True)
                mask = outband.mask.copy()

#                # Write orignal band
#                dst.write_band(i + n_bands_new // 2, outband)

                # Make the values out of range of interest to be nan
                # so that they are ignored in filtering
                if drop_above is not None:
                    mask_above = ma.getdata(outband) > drop_above
                    outband[mask_above] = np.nan
                if drop_below is not None:
                    mask_below = ma.getdata(outband) < drop_below
                    outband[mask_below] = np.nan

                outband_new = generic_filter(
                    outband, LowLevelCallable(nbmean.ctypes), size=size)

                # Mask raster based on result of filter?
                if drop_above is not None:
                    mask_above = ma.getdata(outband) > drop_above
                    mask = np.logical_or(mask, mask_above)
                if drop_below is not None:
                    mask_below = ma.getdata(outband) < drop_below
                    mask = np.logical_or(mask, mask_below)

                outband_new[mask] = dst.nodatavals[bnd_idx]
                outband_ma = ma.masked_array(outband_new, mask=mask)
                dst.write_band(i, outband_ma)


    def generic_filter(self, function, **kwargs: Any) -> None:
        """Apply Gaussian filter to the raster data in-place

        Parameters
        ----------
        function: callable, LowLevelCallable
            Function to be used on the footprint array
        **kwargs : dict, optional
            Keyword arguments passed to SciPy `generic_filter` function

        Returns
        -------
        None
        """

        # TODO: Don't overwrite; add additoinal bands for filtered values

        # NOTE: Adding new bands in this function can result in issues
        # in other parts of the code. Thorough testing is needed for
        # modifying the raster (e.g. hfun add_contour is affected)
        with self.modifying_raster() as dst:
            for i in range(1, self.src.count + 1):
                outband = self.src.read(i, masked=True)
#                # Write orignal band
#                dst.write_band(i + n_bands_new // 2, outband)
                # Write filtered band
                outband = generic_filter(outband, function, **kwargs)
                dst.write_band(i, outband)


    def mask(self,
             shapes: Iterable,
             i: Optional[int] = None,
             **kwargs: Any
             ) -> None:
        """Mask data based on input shapes in-place

        Parameters
        ----------
        shapes : iterable
            List of GeoJSON like dict or objects that implement Python
            geo interface protocol (passed to `rasterio.mask.mask`).
        i : int or None, default=None
            The index of the band to read the data from.
        **kwargs : dict, optional
            Keyword arguments used to create new raster Dataset.

        Returns
        -------
        None
        """

        out_images, out_transform = rasterio.mask.mask(self._src, shapes)
        with self.modifying_raster(**kwargs) as dst:
            if i is None:
                for j in range(1, self.src.count + 1):
                    dst.write_band(j, out_images[j-1])
                    dst.update_tags(j, **self.src.tags(j))
            else:
                for j in range(1, self.src.count + 1):
                    if i == j:
                        dst.write_band(j, out_images[j-1])
                        dst.update_tags(j, **self.src.tags(j))
                    else:
                        dst.write_band(j, self.src.read(j))
                        dst.update_tags(j, **self.src.tags(j))

    def read_masks(self, i: Optional[int] = None) -> npt.NDArray[bool]:
        """Read existing masks on the raster data

        Parameters
        ----------
        i : int or None, default=None

        Returns
        -------
        np.ndarray or view
            Raster band mask from the dataset
        """

        if i is None:
            return np.dstack(
                [self.src.read_masks(i) for i in range(1, self.count + 1)])

        return self.src.read_masks(i)

    def warp(self,
             dst_crs: Union[CRS, str],
             nprocs: int = -1
             ) -> None:
        """Reproject the raster data to specified `dst_crs` in-place

        Parameters
        ----------
        dst_crs : CRS or str
            Destination CRS to which raster must be transformed
        nprocs : int, default=-1
            Number of processors to use for the operation

        Returns
        -------
        None
        """

        nprocs = -1 if nprocs is None else nprocs
        nprocs = multiprocessing.cpu_count() if nprocs == -1 else nprocs
        dst_crs = CRS.from_user_input(dst_crs)
        transform, width, height = warp.calculate_default_transform(
            self.src.crs,
            dst_crs.srs,
            self.src.width,
            self.src.height,
            *self.src.bounds,
            dst_width=self.src.width,
            dst_height=self.src.height
            )

        meta_update = {
            'crs': dst_crs.srs,
            'transform': transform,
            'width': width,
            'height': height
        }
        with self.modifying_raster(**meta_update) as dst:
            for i in range(1, self.src.count + 1):
                rasterio.warp.reproject(
                    source=rasterio.band(self._src, i),
                    destination=rasterio.band(dst, i),
                    src_transform=self.src.transform,
                    crs=self.src.crs,
                    dst_transform=transform,
                    dst_crs=dst_crs.srs,
                    resampling=self.resampling_method,
                    num_threads=nprocs,
                    )


    def resample(self,
                 scaling_factor: float,
                 resampling_method: Optional[str] = None
                 ) -> None:
        """Resample raster data in-place based on a scaling factor

        Parameters
        ----------
        scaling_factor : float
            The scaling factor to use for resampling data
        resampling_method : str or None, default=None
            Name of the resampling method passed to
            `rasterio.DatasetReader.read`

        Returns
        -------
        None
        """

        if resampling_method is None:
            resampling_method = self.resampling_method
        else:
            msg = "resampling_method must be a valid name or None"
            raise ValueError(msg)

        # resample data to target shape
        width = int(self.src.width * scaling_factor)
        height = int(self.src.height * scaling_factor)
        data = self.src.read(
            out_shape=(
                self.src.count,
                height,
                width
            ),
            resampling=resampling_method
        )
        transform = self.src.transform * self.src.transform.scale(
            (self.src.width / data.shape[-1]),
            (self.src.height / data.shape[-2])
        )
        meta_update = {
            'transform': transform,
            'width': width,
            'height': height
            }
        with self.modifying_raster(**meta_update) as dst:
            dst.write(data)

    def save(self, path: Union[str, os.PathLike]) -> None:
        """Save-as raster dataset to a new location

        Parameters
        ----------
        path : str or path-like
            The path to which raster must be saved.

        Returns
        -------
        None
        """

        with rasterio.open(pathlib.Path(path), 'w', **self.src.meta) as dst:
            for i in range(1, self.src.count + 1):
                dst.write_band(i, self.src.read(i))
                dst.update_tags(i, **self.src.tags(i))

    def clip(self, geom: Union[Polygon, MultiPolygon]) -> None:
        """Clip raster data in-place, outside the specified shape.

        Parameters
        ----------
        geom : Polygon or MultiPolygon
            Shape used to clip the raster data

        Returns
        -------
        None
        """

        if isinstance(geom, Polygon):
            geom = MultiPolygon([geom])
        out_image, out_transform = rasterio.mask.mask(
            self.src, geom.geoms, crop=True)
        meta_update = {
            "driver": "GTiff",
            "height": out_image.shape[1],
            "width": out_image.shape[2],
            "transform": out_transform
            }

        with self.modifying_raster(**meta_update) as dest:
            dest.write(out_image)


    def adjust(
            self,
            geom: Union[None, Polygon, MultiPolygon] = None,
            inside_min: float = -np.inf,
            outside_max: float = np.inf,
            cond: Optional[Callable[[npt.NDArray[float]], npt.NDArray[bool]]] = None
            ) -> None:
        """Adjust raster data in-place based on specified shape.

        This method can be used to adjust e.g. raster elevation values
        based on a more accurate land-mass polygon.

        Parameters
        ----------
        geom : None or Polygon or MultiPolygon
            Filled shape to determine which points are considered
            inside or outside (usually land-mass polygon)
        inside_min : float
            The minimum value to truncate raster data that falls
            inside the specified `geom` shape
        outside_max : float
            The maximum value to truncate raster data that falls
            outside the specified `geom` shape

        Returns
        -------
        None
        """

        if isinstance(geom, Polygon):
            geom = MultiPolygon([geom])

        with self.modifying_raster(driver='GTiff') as dst:
            iter_windows = list(self.iter_windows())
            tot = len(iter_windows)
            for i, window in enumerate(iter_windows):
                _logger.debug(f'Processing window {i+1}/{tot}.')

                # NOTE: We should NOT transform polygon, user just
                # needs to make sure input polygon has the same CRS
                # as the hfun (we don't calculate distances in this
                # method)

                start = time()
                values = self.get_values(window=window, masked=True).copy()
                mask = np.zeros_like(values)
                if geom is not None:
                    _logger.info('Creating mask from shape ...')
                    try:
                        mask, _, _ = rasterio.mask.raster_geometry_mask(
                            self.src, geom.geoms,
                            all_touched=True, invert=True)
                        mask = mask[rasterio.windows.window_index(window)]

                    except ValueError:
                        # If there's no overlap between the raster and
                        # shapes then it throws ValueError, instead of
                        # checking for intersection, if there's a value
                        # error we assume there's no overlap
                        _logger.debug(
                            'Polygons don\'t intersect with the raster')

                    _logger.info(
                        f'Creating mask from shape took {time()-start}.')

                if cond:
                    if geom is None:
                        mask = cond(values)
                    else:
                        mask = np.logical_and(mask, cond(values))

                if cond is None and geom is None:
                    raise ValueError(
                        "Neither shape nor condition are provided for adjustment!")

                if mask.any():
                    in_mask = values.mask.copy()
                    values[np.where(np.logical_and(
                            values < inside_min, mask)
                            )] = inside_min
                    values[np.where(np.logical_and(
                            values > outside_max, np.logical_not(mask))
                            )] = outside_max
                    values[in_mask] = dst.nodata
                else:
                    values[values > outside_max] = outside_max

                _logger.info('Write array to file...')
                start = time()
                dst.write_band(1, values, window=window)

                _logger.info(f'Write array to file took {time()-start}.')


    def get_contour(
            self,
            level: float,
            window: Optional[windows.Window] = None
            ) -> Union[LineString, MultiLineString]:
        """Calculate contour lines for specified data level.

        This method can be used e.g. to calculated coastline based on
        raster data.

        Parameters
        ----------
        level : float
            The level for which contour lines must be calculated
        window : windows.Window or None
            The raster window for which contour lines must be calculated

        Returns
        -------
        LineString or MultiLineString
            The contour lines calculated for the specified level
        """

        _logger.debug(
            f'RasterHfun.get_raster_contours(level={level}, window={window})')
        if window is None:
            iter_windows = list(self.iter_windows())
        else:
            iter_windows = [window]
        if len(iter_windows) > 1:
            return self._get_raster_contour_feathered(level, iter_windows)

        return self._get_raster_contour_single_window(level, window)

    def get_channels(
            self,
            level: float = 0,
            width: float = 1000, # in meters
            tolerance: Optional[float] = None
            ) -> Union[Polygon, MultiPolygon]:
        """Calculate narrow width polygons based on specified input

        By using `buffer` functionality this method finds narrow
        regions of the domain. The `level` specifies at which data
        level domain polygon should be calculated and `width`
        describes the narrow region cut-off. `tolerance` is used
        for simplifying the polygon before buffering it to reduce
        computational cost.

        Parameters
        ----------
        level : float, default=0
            Reference level to calculate domain polygon for narrow
            region calculation.
        width : float, default=1000
            Cut-off used for designating narrow regions.
        tolerance : float or None, default=None
            Tolerance used for simplifying domain polygon.

        Returns
        -------
        Polygon or MultiPolygon
            The calculated narrow regions based on raster data
        """

        multipoly = self.get_multipolygon(zmax=level)

        utm_crs = utils.estimate_bounds_utm(
            self.get_bbox().bounds, self.crs)

        if utm_crs is not None:
            transformer = Transformer.from_crs(
                self.src.crs, utm_crs, always_xy=True)
            multipoly = ops.transform(transformer.transform, multipoly)
        channels = utils.get_polygon_channels(
                multipoly, width, simplify=tolerance)
        if channels is None:
            return None

        if utm_crs is not None:
            transformer = Transformer.from_crs(
                utm_crs, self.src.crs, always_xy=True)
            channels = ops.transform(transformer.transform, channels)

        return channels

    def _get_raster_contour_single_window(
            self,
            level: float,
            window: windows.Window
            ) -> Union[LineString, MultiLineString]:
        """Calculate contour on raster data for a single window

        Parameters
        ----------
        level : float
            The level for which contour lines must be calculated
        window : windows.Window or None
            The raster window for which contour lines must be calculated

        Returns
        -------
        LineString or MultiLineString
            The contour lines calculated for the specified level
        """

        x, y = self.get_x(), self.get_y()
        features = []
        values = self.get_values(band=1, window=window)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            _logger.debug('Computing contours...')
            start = time()
            fig, ax = plt.subplots()
            ax.contour(x, y, values, levels=[level])
            _logger.debug(f'Took {time()-start}...')
            plt.close(fig)
        for path_collection in ax.collections:
            for path in path_collection.get_paths():
                try:
                    features.append(LineString(path.vertices))
                except ValueError:
                    # LineStrings must have at least 2 coordinate tuples
                    pass
        return ops.linemerge(features)

    def _get_raster_contour_feathered(
            self,
            level : float,
            iter_windows : Iterable[windows.Window]
            ) -> Union[LineString, MultiLineString]:
        """Wrapper to calculate contour on raster data for a list of windows.

        Parameters
        ----------
        level : float
            The level for which contour lines must be calculated
        iter_windows : iterable
            Sequence of raster windows to calculated the contour lines
            on

        Returns
        -------
        LineString or MultiLineString
            The contour lines calculated for the specified level

        Notes
        -----
        This method calculates contour for each window and then merges
        the results. This private method is a wrapper to the
        method that actually computes the contours.
        """

        with tempfile.TemporaryDirectory(dir=tmpdir) as feather_dir:
            results = self._get_raster_contour_feathered_internal(
                    level, iter_windows, feather_dir)
        return results

    def _get_raster_contour_feathered_internal(
            self,
            level : float,
            iter_windows : Iterable[windows.Window],
            temp_dir : str
            ) -> Union[LineString, MultiLineString]:
        """Calculate contour on raster data for a list of windows.

        Parameters
        ----------
        level : float
            The level for which contour lines must be calculated
        iter_windows : iterable
            Sequence of raster windows to calculated the contour lines
            on
        temp_dir : str
            Path to the temporary directory used for storing feather
            files for each raster window before combining the results

        Returns
        -------
        LineString or MultiLineString
            The contour lines calculated for the specified level

        Notes
        -----
        This method calculates contour for each window and offloads it
        to disk to conserve memory. When all windows are processed,
        `geopandas` out of core method is used to merge the results
        into a `LineString` or `MultiLineString` on the memory
        """

        feathers = []
        total_windows = len(iter_windows)
        _logger.debug(f'Total windows to process: {total_windows}.')
        for i, window in enumerate(iter_windows):
            x, y = self.get_x(window), self.get_y(window)
            _logger.debug(f'Processing window {i+1}/{total_windows}.')
            features = []
            values = self.get_values(band=1, window=window)
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                _logger.debug('Computing contours...')
                start = time()
                fig, ax = plt.subplots()
                ax.contour(x, y, values, levels=[level])
                _logger.debug(f'Took {time()-start}...')
                plt.close(fig)
            for path_collection in ax.collections:
                for path in path_collection.get_paths():
                    try:
                        features.append(LineString(path.vertices))
                    except ValueError:
                        # LineStrings must have at least 2 coordinate tuples
                        pass
            if len(features) > 0:
                tmpfile = os.path.join(temp_dir, f'file_{i}.feather')
                _logger.debug('Saving feather.')
                features = ops.linemerge(features)
                gpd.GeoDataFrame(
                    [{'geometry': features}]
                    ).to_feather(tmpfile)
                feathers.append(tmpfile)
        _logger.debug('Concatenating feathers.')
        features = []
        out = gpd.GeoDataFrame()
        for feather in feathers:
            out = out.append(gpd.read_feather(feather), ignore_index=True)
            feather.unlink()
            geometry = []
            for geom in out.geometry:
                if isinstance(geom, LineString):
                    geometry = MultiLineString([geom])
                    break
            for linestring in geometry:
                features.append(linestring)
        _logger.debug('Merging features.')
        return ops.linemerge(features)

    def iter_windows(
            self,
            chunk_size: Optional[int] = None,
            overlap: Optional[int] = None
            ) -> Generator[windows.Window, None, None]:
        """Calculates sequence of windows for the raster

        This method calculates the sequence of square windows for
        the raster based on the provided `chunk_size` and `overlap`.

        Parameters
        ----------
        chunk_size : int or None , default=None
            Square window size to be used for data chunking
        overlap : int or None , default=None
            Overlap size for calculating chunking windows on the raster

        Yields
        ------
        windows.Window
            Calculated square window on raster based on the window
            size and windows overlap values.
        """

        chunk_size = self.chunk_size if chunk_size is None else chunk_size
        overlap = self.overlap if overlap is None else overlap
        if chunk_size in [0, None, False]:
            yield rasterio.windows.Window(0, 0, self.width, self.height)
            return

        for window in get_iter_windows(
                self.width, self.height, chunk_size, overlap):
            yield window

    def get_window_data(
            self,
            window : windows.Window,
            masked : bool = True,
            band : Optional[int] = None
            ) -> Tuple[npt.NDArray[float], npt.NDArray[float], npt.NDArray[float]]:
        """Return the positions and values of raster data for the window

        Paramaters
        ----------
        window : windows.Window
            The window for which data is to be returned.
        masked : bool, default=True
            Whether to return a masked array in case of missing data.
        band : int or None, default=None
            The ID of the band to read the data from.

        Returns
        -------
        tuple of np.ndarray
            The triple of x-ticks, y-ticks and data on the raster grid
        """

        x0, y0, x1, y1 = self.get_window_bounds(window)
        x = np.linspace(x0, x1, window.width)
        y = np.linspace(y1, y0, window.height)
        if band is not None:
            data = self.src.read(band, masked=masked, window=window)
        else:
            data = self.src.read(masked=masked, window=window)
        return x, y, data

    def get_window_bounds(
            self,
            window : windows.Window
            ) -> Tuple[int, int, int, int]:
        """Returns west, south, east, north bounds of the window

        Paramaters
        ----------
        window : windows.Window
            The window for which bounds are calculated

        Returns
        -------
        tuple of int
            West, south, east, north bounds of the window
        """

        return array_bounds(
            window.height,
            window.width,
            self.get_window_transform(window))

    def get_window_transform(
            self,
            window : windows.Window
            ) -> Union[None, Affine]:
        """Returns raster's affine transform calculated for the window

        Paramaters
        ----------
        window : windows.Window
            The window for which the affine transform is calculated

        Returns
        -------
        Affine
            Affine transform matrix for specified window
        """

        if window is None:
            return None
        return windows.transform(window, self.transform)

    @property
    def x(self) -> npt.NDArray[float]:
        """Read-only attribute for the x-ticks of raster grid

        This is a read-only property that returns the same results as
        `get_x` method
        """

        return self.get_x()

    @property
    def y(self) -> npt.NDArray[float]:
        """Read-only attribute for the y-ticks of raster grid

        This is a read-only property that returns the same results as
        `get_y` method
        """

        return self.get_y()

    @property
    def values(self) -> npt.NDArray[float]:
        """Read-only attribute for the raster grid data

        This is a read-only property that returns the same results as
        `get_values` method
        """

        return self.get_values()

    @property
    def path(self) -> pathlib.Path:
        """Read-only attribute for the path to original raster file"""

        return self._path

    @property
    def tmpfile(self) -> pathlib.Path:
        """Read-only attribute for the path to working raster file"""

        return self._tmpfile

    @property
    def md5(self) -> str:
        """Read-only attribute for the hash of working raster file content

        This is a read-only property that is recalculated every time
        from the content of the temporary working raster file.
        """

        hash_md5 = hashlib.md5()
        with open(self._tmpfile.resolve(), "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    @property
    def count(self) -> int:
        """Read-only attribute for the number of bands of raster dataset"""

        return self.src.count

    @property
    def is_masked(self) -> bool:
        """Read-only attribute indicating whether raster has missing data.

        This is a read-only property that indicates whether or not
        the raster has missing data points. The value of this
        property is recalculated every time the property is retrieved.
        """

        for window in self.iter_windows(self.chunk_size):
            if self.src.nodata in self.src.read(window=window):
                return True
        return False

    @property
    def shape(self) -> Tuple[int, int]:
        """Read-only attribute indicating the shape of raster grid"""

        return self.src.shape

    @property
    def height(self) -> int:
        """Read-only attribute for the number of rows of raster grid"""

        return self.src.height

    @property
    def bbox(self) -> Union[Polygon, Bbox]:
        """Read-only attribute for the bounding box of the raster grid

        This is a read-only property that returns the same results as
        `get_bbox` method
        """

        return self.get_bbox()

    @property
    def src(self) -> rasterio.DatasetReader:
        """Read-only attribute for access to opened dataset handle"""

        return self._src

    @property
    def width(self) -> int:
        """Read-only attribute for the number of columns of raster grid"""

        return self.src.width

    @property
    def dx(self) -> float:
        """Read-only attribute for grid distance in x direction"""
        return self.src.transform[0]

    @property
    def dy(self) -> float:
        """Read-only attribute for grid distance in y direction"""
        return -self.src.transform[4]

    @property
    def crs(self) -> CRS:
        """Read-only attribute for raster CRS

        This read-only property returns CRS as a pyproj.CRS type
        **not** rasterio.CRS.
        """

        # cast rasterio.CRS to pyproj.CRS for API consistency
        return CRS.from_user_input(self.src.crs)

    @property
    def nodatavals(self) -> float:
        return self.src.nodatavals

    @property
    def transform(self) -> Affine:
        """Read-only attribute for raster's transform"""
        return self.src.transform

    @property
    def dtypes(self) -> Any:
        """Read-only attribute for raster's data type"""
        return self.src.dtypes

    @property
    def nodata(self):
        return self.src.nodata

    @property
    def xres(self) -> float:
        """Read-only attribute for grid resolution in x direction

        This read-only property returns the same value as `dx`
        """
        return self.transform[0]

    @property
    def yres(self) -> float:
        """Read-only attribute for grid resolution in y direction

        This read-only property returns the same value as `-dy`
        """
        return self.transform[4]

    @property
    def resampling_method(self) -> Resampling:
        """Modifiable attribute for stored raster resampling method"""

        if not hasattr(self, '_resampling_method'):
            self._resampling_method = Resampling.nearest
        return self._resampling_method

    @resampling_method.setter
    def resampling_method(self, resampling_method: Resampling) -> None:
        """Set `resampling_method`"""

        if not isinstance(resampling_method, Resampling):
            TypeError(
                f'Argument resampling_method must be of type  {Resampling}, '
                f'not type {type(resampling_method)}.')
        self._resampling_method = resampling_method

    @property
    def chunk_size(self) -> int:
        """Modfiable attribute for stored square raster window size"""
        return self._chunk_size

    @chunk_size.setter
    def chunk_size(self, chunk_size: int) -> None:
        """Set `chunk_size`"""

        self._chunk_size = chunk_size

    @property
    def overlap(self) -> int:
        """Modfiable attribute for stored raster windows overlap amount"""
        return self._overlap

    @overlap.setter
    def overlap(self, overlap: int) -> None:
        """Set `overlap`"""

        self._overlap = overlap


def get_iter_windows(
        width: int,
        height: int,
        chunk_size: int = 0,
        overlap: int = 0,
        row_off: int = 0,
        col_off: int = 0
        ) -> Generator[windows.Window, None, None]:
    """Calculates sequence of raster windows based on basic inputs

    This stand-alone function calculates the sequence of square windows
    based on inputs that are not necessarily tied to a specific raster
    dataset.

    Parameters
    ----------
    width : int
        The total width of the all unioned windows combined
    height : int
        The total height of the all unioned windows combined
    chunk_size : int, default=0
        Square window size to be used for data chunking
    overlap: int, default=0
        Overlap size for calculating chunking windows on the raster
    row_off: int, default=0
        Unused!
    col_off: int, default=0
        Unused!

    Yields
    ------
    windows.Window
        Calculated square window based on the total size as well as
        chunk size and windows overlap values.
    """

    win_h = chunk_size + overlap
    win_w = chunk_size + overlap
    n_win_h = math.ceil(height / chunk_size)
    n_win_w = math.ceil(width / chunk_size)
    for i in range(n_win_h):
        for j in range(n_win_w):
            off_h = i * chunk_size
            off_w = j * chunk_size
            h = chunk_size + overlap
            h = h - (off_h + h) % height if off_h + h > height else h
            w = chunk_size + overlap
            w = w - (off_w + w) % width if off_w + w > width else w
            yield windows.Window(off_w, off_h, w, h)


def redistribute_vertices(
        geom: Union[LineString, MultiLineString],
        distance: float
        ) -> Union[LineString, MultiLineString]:
    """Redistribute the vertices of input line strings

    Parameters
    ----------
    geom : LineString or MultiLineString
        Input line strings whose vertices is to be redistributed.
    distance : float
        The distance to be used for redistribution.

    Returns
    -------
    LineString or MultiLineString
        The resulting line strings with redistributed vertices.

    Raises
    ------
    ValueError
        If input geometry is not LineString or MultiLineString.
    """

    if geom.geom_type == 'LineString': # pylint: disable=R1705
        num_vert = int(round(geom.length / distance))
        if num_vert == 0:
            num_vert = 1
        return LineString(
            [geom.interpolate(float(n) / num_vert, normalized=True)
             for n in range(num_vert + 1)])
    elif geom.geom_type == 'MultiLineString':
        parts = [redistribute_vertices(part, distance)
                 for part in geom]
        return type(geom)([p for p in parts if not p.is_empty])

    raise ValueError(f'unhandled geometry {geom.geom_type}')
